# -*- coding: utf-8 -*-
"""
reV bespoke wind plant analysis tools
"""

# pylint: disable=anomalous-backslash-in-string
import copy
import json
import logging
import os
import time
from concurrent.futures import as_completed
from importlib import import_module
from inspect import signature
from numbers import Number
from warnings import warn

import numpy as np
import pandas as pd
import psutil
from rex.joint_pd.joint_pd import JointPD
from rex.multi_year_resource import MultiYearWindResource
from rex.utilities.bc_parse_table import parse_bc_table
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import create_dirs, log_mem
from rex.utilities.utilities import parse_year

from reV.config.output_request import SAMOutputRequest
from reV.econ.utilities import lcoe_fcr
from reV.generation.generation import Gen
from reV.handlers.exclusions import ExclusionLayers
from reV.handlers.outputs import Outputs
from reV.SAM.generation import WindPower, WindPowerPD
from reV.supply_curve.aggregation import AggFileHandler, BaseAggregation
from reV.supply_curve.extent import SupplyCurveExtent
from reV.supply_curve.points import AggregationSupplyCurvePoint as AggSCPoint
from reV.supply_curve.points import SupplyCurvePoint
from reV.utilities import (
    ModuleName,
    ResourceMetaField,
    SupplyCurveField,
    log_versions,
)
from reV.utilities.exceptions import EmptySupplyCurvePointError, FileInputError

logger = logging.getLogger(__name__)


class BespokeMultiPlantData:
    """Multi-plant preloaded data.

    This object is intended to facilitate the use of pre-loaded data for
    running :class:`BespokeWindPlants` on systems with slow parallel
    reads to a single HDF5 file.
    """

    def __init__(self, res_fpath, sc_gid_to_hh, sc_gid_to_res_gid,
                 pre_load_humidity=False):
        """Initialize BespokeMultiPlantData

        Parameters
        ----------
        res_fpath : str | list
            Unix shell style path (potentially containing wildcard (*)
            patterns) to a single or multi-file resource file set(s).
            Can also be an explicit list of resource file paths, which
            themselves can contain wildcards. This input must be
            readable by
            :py:class:`rex.multi_year_resource.MultiYearWindResource`.
        sc_gid_to_hh : dict
            Dictionary mapping SC GID values to hub-heights. Data for
            each SC GID will be pulled for the corresponding hub-height
            given in this dictionary.
        sc_gid_to_res_gid : dict
            Dictionary mapping SC GID values to an iterable oif resource
            GID values. Resource GID values should correspond to GID
            values in the HDF5 file, so any GID map must be applied
            before initializing :class`BespokeMultiPlantData`.
        pre_load_humidity : optional, default=False
            Option to pre-load relative humidity data (useful for icing
            runs). If ``False``, relative humidities are not loaded.
        """
        self.res_fpath = res_fpath
        self.sc_gid_to_hh = sc_gid_to_hh
        self.sc_gid_to_res_gid = sc_gid_to_res_gid
        self.hh_to_res_gids = {}
        self._wind_dirs = None
        self._wind_speeds = None
        self._temps = None
        self._pressures = None
        self._relative_humidities = None
        self._pre_load_humidity = pre_load_humidity
        self._time_index = None
        self._pre_load_data()

    def _pre_load_data(self):
        """Pre-load the resource data."""

        for sc_gid, gids in self.sc_gid_to_res_gid.items():
            hh = self.sc_gid_to_hh[sc_gid]
            self.hh_to_res_gids.setdefault(hh, set()).update(gids)

        self.hh_to_res_gids = {
            hh: sorted(gids) for hh, gids in self.hh_to_res_gids.items()
        }

        start_time = time.time()
        with MultiYearWindResource(self.res_fpath) as res:
            self._wind_dirs = {
                hh: res[f"winddirection_{hh}m", :, gids]
                for hh, gids in self.hh_to_res_gids.items()
            }
            self._wind_speeds = {
                hh: res[f"windspeed_{hh}m", :, gids]
                for hh, gids in self.hh_to_res_gids.items()
            }
            self._temps = {
                hh: res[f"temperature_{hh}m", :, gids]
                for hh, gids in self.hh_to_res_gids.items()
            }
            self._pressures = {
                hh: res[f"pressure_{hh}m", :, gids]
                for hh, gids in self.hh_to_res_gids.items()
            }
            self._time_index = res.time_index
            if self._pre_load_humidity:
                self._relative_humidities = {
                    hh: res["relativehumidity_2m", :, gids]
                    for hh, gids in self.hh_to_res_gids.items()
                }

        logger.debug(
            f"Data took {(time.time() - start_time) / 60:.2f} " f"min to load"
        )

    def get_preloaded_data_for_gid(self, sc_gid):
        """Get the pre-loaded data for a single SC GID.

        Parameters
        ----------
        sc_gid : int
            SC GID to load resource data for.

        Returns
        -------
        BespokeSinglePlantData
            A loaded ``BespokeSinglePlantData`` object that can act as
            an HDF5 handler stand-in *for this SC GID only*.
        """
        hh = self.sc_gid_to_hh[sc_gid]
        sc_point_res_gids = sorted(self.sc_gid_to_res_gid[sc_gid])
        data_inds = np.searchsorted(self.hh_to_res_gids[hh], sc_point_res_gids)

        rh = (None if not self._pre_load_humidity
              else self._relative_humidities[hh][:, data_inds])
        return BespokeSinglePlantData(
            sc_point_res_gids,
            self._wind_dirs[hh][:, data_inds],
            self._wind_speeds[hh][:, data_inds],
            self._temps[hh][:, data_inds],
            self._pressures[hh][:, data_inds],
            self._time_index,
            rh,
        )


class BespokeSinglePlantData:
    """Single-plant preloaded data.

    This object is intended to facilitate the use of pre-loaded data for
    running :class:`BespokeSinglePlant` on systems with slow parallel
    reads to a single HDF5 file.
    """

    def __init__(
        self, data_inds, wind_dirs, wind_speeds, temps, pressures, time_index,
        relative_humidities=None,
    ):
        """Initialize BespokeSinglePlantData

        Parameters
        ----------
        data_inds : 1D np.array
            Array of res GIDs. This array should be the same length as
            the second dimension of `wind_dirs`, `wind_speeds`, `temps`,
            and `pressures`. The GID value of data_inds[0] should
            correspond to the `wind_dirs[:, 0]` data, etc.
        wind_dirs : 2D np.array
            Array of wind directions. Dimensions should be correspond to
            [time, location]. See documentation for `data_inds` for
            required spatial mapping of GID values.
        wind_speeds : 2D np.array
            Array of wind speeds. Dimensions should be correspond to
            [time, location]. See documentation for `data_inds` for
            required spatial mapping of GID values.
        temps : 2D np.array
            Array oftemperatures. Dimensions should be correspond to
            [time, location]. See documentation for `data_inds` for
            required spatial mapping of GID values.
        pressures : 2D np.array
            Array of pressures. Dimensions should be correspond to
            pressures, respectively. Dimensions should be correspond to
            [time, location]. See documentation for `data_inds` for
            required spatial mapping of GID values.
        time_index : 1D np.array
            Time index array corresponding to the temporal dimension of
            the 2D data. Will be exposed directly to user.
        relative_humidities : 2D np.array, optional
            Array of relative humidities. Dimensions should be
            correspond to [time, location]. See documentation for
            `data_inds` for required spatial mapping of GID values.
            If ``None``, relative_humidities cannot be queried.
        """

        self.data_inds = data_inds
        self.wind_dirs = wind_dirs
        self.wind_speeds = wind_speeds
        self.temps = temps
        self.pressures = pressures
        self.time_index = time_index
        self.relative_humidities = relative_humidities
        self._humidities_exist = relative_humidities is not None

    def __getitem__(self, key):
        dset_name, t_idx, gids = key
        data_inds = np.searchsorted(self.data_inds, gids)
        if "winddirection" in dset_name:
            return self.wind_dirs[t_idx, data_inds]
        if "windspeed" in dset_name:
            return self.wind_speeds[t_idx, data_inds]
        if "temperature" in dset_name:
            return self.temps[t_idx, data_inds]
        if "pressure" in dset_name:
            return self.pressures[t_idx, data_inds]
        if self._humidities_exist and "relativehumidity" in dset_name:
            return self.relative_humidities[t_idx, data_inds]
        msg = f"Unknown dataset name: {dset_name!r}"
        logger.error(msg)
        raise ValueError(msg)


class BespokeSinglePlant:
    """Framework for analyzing and optimizing a wind plant layout specific to
    the local wind resource and exclusions for a single reV supply curve point.
    """

    DEPENDENCIES = ("shapely",)
    OUT_ATTRS = copy.deepcopy(Gen.OUT_ATTRS)

    def __init__(
        self,
        gid,
        excl,
        res,
        tm_dset,
        sam_sys_inputs,
        objective_function,
        capital_cost_function,
        fixed_operating_cost_function,
        variable_operating_cost_function,
        balance_of_system_cost_function,
        min_spacing="5x",
        wake_loss_multiplier=1,
        ga_kwargs=None,
        output_request=("system_capacity", "cf_mean"),
        ws_bins=(0.0, 20.0, 5.0),
        wd_bins=(0.0, 360.0, 45.0),
        excl_dict=None,
        inclusion_mask=None,
        data_layers=None,
        resolution=64,
        excl_area=None,
        exclusion_shape=None,
        eos_mult_baseline_cap_mw=200,
        prior_meta=None,
        gid_map=None,
        bias_correct=None,
        pre_loaded_data=None,
        close=True,
    ):
        """
        Parameters
        ----------
        gid : int
            gid for supply curve point to analyze.
        excl : str | ExclusionMask
            Filepath to exclusions h5 or ExclusionMask file handler.
        res : str | Resource
            Filepath to .h5 wind resource file or pre-initialized Resource
            handler
        tm_dset : str
            Dataset name in the exclusions file containing the
            exclusions-to-resource mapping data.
        sam_sys_inputs : dict
            SAM windpower compute module system inputs not including the
            wind resource data.
        objective_function : str
            The objective function of the optimization as a string, should
            return the objective to be minimized during layout optimization.
            Variables available are:

                - ``n_turbines``: the number of turbines
                - ``system_capacity``: wind plant capacity
                - ``aep``: annual energy production
                - ``avg_sl_dist_to_center_m``: Average straight-line
                  distance to the supply curve point center from all
                  turbine locations (in m). Useful for computing plant
                  BOS costs.
                - ``avg_sl_dist_to_medoid_m``: Average straight-line
                  distance to the medoid of all turbine locations
                  (in m). Useful for computing plant BOS costs.
                - ``nn_conn_dist_m``: Total BOS connection distance
                  using nearest-neighbor connections. This variable is
                  only available for the
                  ``balance_of_system_cost_function`` equation.
                - ``fixed_charge_rate``: user input fixed_charge_rate if
                  included as part of the sam system config.
                - ``capital_cost``: plant capital cost as evaluated
                  by `capital_cost_function`
                - ``fixed_operating_cost``: plant fixed annual operating
                  cost as evaluated by `fixed_operating_cost_function`
                - ``variable_operating_cost``: plant variable annual
                  operating cost as evaluated by
                  `variable_operating_cost_function`
                - ``balance_of_system_cost``: plant balance of system
                  cost as evaluated by `balance_of_system_cost_function`
                - ``self.wind_plant``: the SAM wind plant object,
                  through which all SAM variables can be accessed

        capital_cost_function : str
            The plant capital cost function as a string, must return the total
            capital cost in $. Has access to the same variables as the
            objective_function.
        fixed_operating_cost_function : str
            The plant annual fixed operating cost function as a string, must
            return the fixed operating cost in $/year. Has access to the same
            variables as the objective_function.
        variable_operating_cost_function : str
            The plant annual variable operating cost function as a string, must
            return the variable operating cost in $/kWh. Has access to the same
            variables as the objective_function. You can set this to "0"
            to effectively ignore variable operating costs.
        balance_of_system_cost_function : str
            The plant balance-of-system cost function as a string, must
            return the variable operating cost in $. Has access to the
            same variables as the objective_function. You can set this
            to "0" to effectively ignore balance-of-system costs.
        balance_of_system_cost_function : str
            The plant balance-of-system cost function as a string, must
            return the variable operating cost in $. Has access to the same
            variables as the objective_function.
        min_spacing : float | int | str
            Minimum spacing between turbines in meters. Can also be a string
            like "5x" (default) which is interpreted as 5 times the turbine
            rotor diameter.
        wake_loss_multiplier : float, optional
            A multiplier used to scale the annual energy lost due to
            wake losses.
            .. WARNING:: This multiplier will ONLY be applied during the
            optimization process and will NOT be come through in output
            values such as the hourly profiles,
            aep, any of the cost functions, or even the output objective.
        ga_kwargs : dict | None
            Dictionary of keyword arguments to pass to GA initialization.
            If `None`, default initialization values are used.
            See :class:`~reV.bespoke.gradient_free.GeneticAlgorithm` for
            a description of the allowed keyword arguments.
        output_request : list | tuple
            Outputs requested from the SAM windpower simulation after the
            bespoke plant layout optimization. Can also request resource means
            like ws_mean, windspeed_mean, temperature_mean, pressure_mean.
        ws_bins : tuple
            3-entry tuple with (start, stop, step) for the windspeed binning of
            the wind joint probability distribution. The stop value is
            inclusive, so ws_bins=(0, 20, 5) would result in four bins with bin
            edges (0, 5, 10, 15, 20).
        wd_bins : tuple
            3-entry tuple with (start, stop, step) for the winddirection
            binning of the wind joint probability distribution. The stop value
            is inclusive, so ws_bins=(0, 360, 90) would result in four bins
            with bin edges (0, 90, 180, 270, 360).
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
            None if excl input is pre-initialized.
        inclusion_mask : np.ndarray
            2D array pre-extracted inclusion mask where 1 is included and 0 is
            excluded. The shape of this will be checked against the input
            resolution.
        data_layers : None | dict
            Aggregation data layers. Must be a dictionary keyed by data label
            name. Each value must be another dictionary with "dset", "method",
            and "fpath".
        resolution : int
            Number of exclusion points per SC point along an axis.
            This number**2 is the total number of exclusion points per
            SC point.
        excl_area : float | None, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath, by default None
        exclusion_shape : tuple
            Shape of the full exclusions extent (rows, cols). Inputing this
            will speed things up considerably.
        eos_mult_baseline_cap_mw : int | float, optional
            Baseline plant capacity (MW) used to calculate economies of
            scale (EOS) multiplier from the `capital_cost_function`. EOS
            multiplier is calculated as the $-per-kW of the wind plant
            divided by the $-per-kW of a plant with this baseline
            capacity. By default, `200` (MW), which aligns the baseline
            with ATB assumptions. See here: https://tinyurl.com/y85hnu6h.
        prior_meta : pd.DataFrame | None
            Optional meta dataframe belonging to a prior run. This will only
            run the timeseries power generation step and assume that all of the
            wind plant layouts are fixed given the prior run. The meta data
            needs columns "capacity", "turbine_x_coords", and
            "turbine_y_coords".
        gid_map : None | str | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This can be None, a pre-extracted dict, or
            a filepath to json or csv. If this is a csv, it must have the
            columns "gid" (which matches the techmap) and "gid_map" (gids to
            extract from the resource input). This is useful if you're running
            forecasted resource data (e.g., ECMWF) to complement historical
            meteorology (e.g., WTK).
        bias_correct : str | pd.DataFrame, optional
            Optional DataFrame or CSV filepath to a wind or solar
            resource bias correction table. This has columns:

                - ``gid``: GID of site (can be index name of dataframe)
                - ``method``: function name from ``rex.bias_correction`` module

            The ``gid`` field should match the true resource ``gid`` regardless
            of the optional ``gid_map`` input. Only ``windspeed`` **or**
            ``GHI`` + ``DNI`` + ``DHI`` are corrected, depending on the
            technology (wind for the former, PV or CSP for the latter). See the
            functions in the ``rex.bias_correction`` module for available
            inputs for ``method``. Any additional kwargs required for the
            requested ``method`` can be input as additional columns in the
            ``bias_correct`` table e.g., for linear bias correction functions
            you can include ``scalar`` and ``adder`` inputs as columns in the
            ``bias_correct`` table on a site-by-site basis. If ``None``, no
            corrections are applied. By default, ``None``.
        pre_loaded_data : BespokeSinglePlantData, optional
            A pre-loaded :class:`BespokeSinglePlantData` object, or
            ``None``. Can be useful to speed up execution on file
            systems with slow parallel reads.
        close : bool
            Flag to close object file handlers on exit.
        """

        logger.debug(
            "Initializing BespokeSinglePlant for gid {}...".format(gid)
        )
        logger.debug("Resource filepath: {}".format(res))
        logger.debug("Exclusion filepath: {}".format(excl))
        logger.debug("Exclusion dict: {}".format(excl_dict))
        logger.debug(
            "Bespoke objective function: {}".format(objective_function)
        )
        logger.debug("Bespoke cost function: {}".format(objective_function))
        logger.debug(
            "Bespoke wake loss multiplier: {}".format(wake_loss_multiplier)
        )
        logger.debug("Bespoke GA initialization kwargs: {}".format(ga_kwargs))
        logger.debug(
            "Bespoke EOS multiplier baseline capacity: {:,} MW".format(
                eos_mult_baseline_cap_mw
            )
        )

        if isinstance(min_spacing, str) and min_spacing.endswith("x"):
            rotor_diameter = sam_sys_inputs["wind_turbine_rotor_diameter"]
            min_spacing = float(min_spacing.strip("x")) * rotor_diameter

        if not isinstance(min_spacing, (int, float)):
            try:
                min_spacing = float(min_spacing)
            except Exception as e:
                msg = (
                    "min_spacing must be numeric but received: {}, {}".format(
                        min_spacing, type(min_spacing)
                    )
                )
                logger.error(msg)
                raise TypeError(msg) from e

        self.objective_function = objective_function
        self.capital_cost_function = capital_cost_function
        self.fixed_operating_cost_function = fixed_operating_cost_function
        self.variable_operating_cost_function = (
            variable_operating_cost_function
        )
        self.balance_of_system_cost_function = balance_of_system_cost_function
        self.min_spacing = min_spacing
        self.wake_loss_multiplier = wake_loss_multiplier
        self.ga_kwargs = ga_kwargs or {}

        self._sam_sys_inputs = sam_sys_inputs
        self._out_req = list(output_request)
        self._ws_bins = ws_bins
        self._wd_bins = wd_bins
        self._baseline_cap_mw = eos_mult_baseline_cap_mw

        self._res_df = None
        self._prior_meta = prior_meta is not None
        self._meta = prior_meta
        self._wind_dist = None
        self._ws_edges = None
        self._wd_edges = None
        self._wind_plant_pd = None
        self._wind_plant_ts = None
        self._plant_optm = None
        self._gid_map = self._parse_gid_map(gid_map)
        self._bias_correct = Gen._parse_bc(bias_correct)
        self._pre_loaded_data = pre_loaded_data
        self._outputs = {}

        res = res if not isinstance(res, str) else MultiYearWindResource(res)

        self._sc_point = AggSCPoint(
            gid,
            excl,
            res,
            tm_dset,
            excl_dict=excl_dict,
            inclusion_mask=inclusion_mask,
            resolution=resolution,
            excl_area=excl_area,
            exclusion_shape=exclusion_shape,
            close=close,
        )

        self._parse_output_req()
        self._data_layers = data_layers
        self._parse_prior_run()

    def __str__(self):
        s = "BespokeSinglePlant for reV SC gid {} with resolution {}".format(
            self.sc_point.gid, self.sc_point.resolution
        )
        return s

    def __repr__(self):
        s = "BespokeSinglePlant for reV SC gid {} with resolution {}".format(
            self.sc_point.gid, self.sc_point.resolution
        )
        return s

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        if type is not None:
            raise

    def _parse_output_req(self):
        """Make sure that the output request has basic important parameters
        (cf_mean, annual_energy) and process mean wind resource datasets
        (ws_mean, *_mean) if requested.
        """

        required = ("cf_mean", "annual_energy")
        for req in required:
            if req not in self._out_req:
                self._out_req.append(req)

        if "ws_mean" in self._out_req:
            self._out_req.remove("ws_mean")
            self._outputs["ws_mean"] = self.res_df["windspeed"].mean()

        for req in copy.deepcopy(self._out_req):
            if req in self.res_df:
                self._out_req.remove(req)
                for annual_ti in self.annual_time_indexes:
                    year = annual_ti.year[0]
                    mask = self.res_df.index.isin(annual_ti)
                    arr = self.res_df.loc[mask, req].values.flatten()
                    self._outputs[req + f"-{year}"] = arr

            elif req.replace("_mean", "") in self.res_df:
                self._out_req.remove(req)
                dset = req.replace("_mean", "")
                self._outputs[req] = self.res_df[dset].mean()

        if "lcoe_fcr" in self._out_req and (
            "fixed_charge_rate" not in self.original_sam_sys_inputs
        ):
            msg = (
                'User requested "lcoe_fcr" but did not input '
                '"fixed_charge_rate" in the SAM system config.'
            )
            logger.error(msg)
            raise KeyError(msg)

    def _parse_prior_run(self):
        """Parse prior bespoke wind plant optimization run meta data and make
        sure the SAM system inputs are set accordingly."""

        # {meta_column: sam_sys_input_key}
        required = {
            SupplyCurveField.CAPACITY_AC_MW: "system_capacity",
            SupplyCurveField.TURBINE_X_COORDS: "wind_farm_xCoordinates",
            SupplyCurveField.TURBINE_Y_COORDS: "wind_farm_yCoordinates",
        }

        if self._prior_meta:
            missing = [k for k in required if k not in self.meta]
            msg = (
                "Prior bespoke run meta data is missing the following "
                "required columns: {}".format(missing)
            )
            assert not any(missing), msg

            for meta_col, sam_sys_key in required.items():
                prior_value = self.meta[meta_col].values[0]
                self._sam_sys_inputs[sam_sys_key] = prior_value

            # convert reV supply curve cap in MW to SAM capacity in kW
            self._sam_sys_inputs["system_capacity"] *= 1e3

    @staticmethod
    def _parse_gid_map(gid_map):
        """Parse the gid map and return the extracted dictionary or None if not
        provided

        Parameters
        ----------
        gid_map : None | str | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This can be None, a pre-extracted dict, or
            a filepath to json or csv. If this is a csv, it must have the
            columns "gid" (which matches the techmap) and "gid_map" (gids to
            extract from the resource input). This is useful if you're running
            forecasted resource data (e.g., ECMWF) to complement historical
            meteorology (e.g., WTK).

        Returns
        -------
        gid_map : dict | None
            Pre-extracted gid_map dictionary if provided or None if not.
        """

        if isinstance(gid_map, str):
            if gid_map.endswith(".csv"):
                gid_map = (
                    pd.read_csv(gid_map)
                    .rename(SupplyCurveField.map_to(ResourceMetaField), axis=1)
                    .to_dict()
                )
                err_msg = f"Need {ResourceMetaField.GID} in gid_map column"
                assert ResourceMetaField.GID in gid_map, err_msg
                assert "gid_map" in gid_map, 'Need "gid_map" in gid_map column'
                gid_map = {
                    gid_map[ResourceMetaField.GID][i]: gid_map["gid_map"][i]
                    for i in gid_map[ResourceMetaField.GID]
                }

            elif gid_map.endswith(".json"):
                with open(gid_map) as f:
                    gid_map = json.load(f)

        return gid_map

    def close(self):
        """Close any open file handlers via the sc point attribute. If this
        class was initialized with close=False, this will not close any
        handlers."""
        self.sc_point.close()

    def bias_correct_ws(self, ws, dset, h5_gids):
        """Bias correct windspeed data if the ``bias_correct`` input was
        provided.

        Parameters
        ----------
        ws : np.ndarray
            Windspeed data in shape (time, space)
        dset : str
            Resource dataset name e.g., "windspeed_100m", "temperature_100m",
            "pressure_100m", or something similar
        h5_gids : list | np.ndarray
            Array of integer gids (spatial indices) from the source h5 file.
            This is used to get the correct bias correction parameters from
            ``bias_correct`` table based on its ``gid`` column

        Returns
        -------
        ws : np.ndarray
            Bias corrected windspeed data in same shape as input
        """

        if self._bias_correct is not None and dset.startswith("windspeed_"):
            out = parse_bc_table(self._bias_correct, h5_gids)
            bc_fun, bc_fun_kwargs, bool_bc = out

            if bool_bc.any():
                logger.debug(
                    "Bias correcting windspeed with function {} "
                    "for h5 gids: {}".format(bc_fun, h5_gids)
                )

                bc_fun_kwargs["ws"] = ws[:, bool_bc]
                sig = signature(bc_fun)
                bc_fun_kwargs = {
                    k: v
                    for k, v in bc_fun_kwargs.items()
                    if k in sig.parameters
                }

                ws[:, bool_bc] = bc_fun(**bc_fun_kwargs)

        return ws

    def get_weighted_res_ts(self, dset):
        """Special method for calculating the exclusion-weighted mean resource
        timeseries data for the BespokeSinglePlant.

        Parameters
        ----------
        dset : str
            Resource dataset name e.g., "windspeed_100m", "temperature_100m",
            "pressure_100m", or something similar

        Returns
        -------
        data : np.ndarray
            Timeseries data of shape (n_time,) for the wind plant weighted by
            the plant inclusions mask.
        """
        gids = self.sc_point.h5_gid_set
        h5_gids = copy.deepcopy(gids)
        if self._gid_map is not None:
            h5_gids = [self._gid_map[g] for g in gids]

        if self._pre_loaded_data is None:
            data = self.sc_point.h5[dset, :, h5_gids]
        else:
            data = self._pre_loaded_data[dset, :, h5_gids]

        data = self.bias_correct_ws(data, dset, h5_gids)

        weights = np.zeros(len(gids))
        for i, gid in enumerate(gids):
            mask = self.sc_point._h5_gids == gid
            weights[i] = self.sc_point.include_mask_flat[mask].sum()

        weights /= weights.sum()
        data = data.astype(np.float32)
        data *= weights
        data = np.sum(data, axis=1)

        return data

    def get_weighted_res_dir(self):
        """Special method for calculating the exclusion-weighted mean wind
        direction for the BespokeSinglePlant

        Returns
        -------
        mean_wind_dirs : np.ndarray
            Timeseries array of winddirection data in shape (n_time,) in units
            of degrees from north.
        """

        dset = f"winddirection_{self.hub_height}m"
        gids = self.sc_point.h5_gid_set
        h5_gids = copy.deepcopy(gids)
        if self._gid_map is not None:
            h5_gids = [self._gid_map[g] for g in gids]

        if self._pre_loaded_data is None:
            dirs = self.sc_point.h5[dset, :, h5_gids]
        else:
            dirs = self._pre_loaded_data[dset, :, h5_gids]
        angles = np.radians(dirs, dtype=np.float32)

        weights = np.zeros(len(gids))
        for i, gid in enumerate(gids):
            mask = self.sc_point._h5_gids == gid
            weights[i] = self.sc_point.include_mask_flat[mask].sum()

        weights /= weights.sum()
        sin = np.sum(np.sin(angles) * weights, axis=1)
        cos = np.sum(np.cos(angles) * weights, axis=1)

        mean_wind_dirs = np.degrees(np.arctan2(sin, cos))
        mean_wind_dirs[(mean_wind_dirs < 0)] += 360

        return mean_wind_dirs

    @property
    def gid(self):
        """SC point gid for this bespoke plant.

        Returns
        -------
        int
        """
        return self.sc_point.gid

    @property
    def include_mask(self):
        """Get the supply curve point 2D inclusion mask (included is 1,
        excluded is 0)

        Returns
        -------
        np.ndarray
        """
        return self.sc_point.include_mask

    @property
    def pixel_side_length(self):
        """Get the length of a single exclusion pixel side (meters)

        Returns
        -------
        float
        """
        return np.sqrt(self.sc_point.pixel_area) * 1000.0

    @property
    def original_sam_sys_inputs(self):
        """Get the original (pre-optimized) SAM windpower system inputs.

        Returns
        -------
        dict
        """
        return self._sam_sys_inputs

    @property
    def sam_sys_inputs(self):
        """Get the SAM windpower system inputs. If the wind plant has not yet
        been optimized, this returns the initial SAM config. If the wind plant
        has been optimized using the wind_plant_pd object, this returns the
        final optimized SAM plant config.

        Returns
        -------
        dict
        """
        config = copy.deepcopy(self._sam_sys_inputs)
        if self._wind_plant_pd is None:
            return config

        config.update(self._wind_plant_pd.sam_sys_inputs)
        return config

    @property
    def sc_point(self):
        """Get the reV supply curve point object.

        Returns
        -------
        AggSCPoint
        """
        return self._sc_point

    @property
    def meta(self):
        """Get the basic supply curve point meta data

        Returns
        -------
        pd.DataFrame
        """
        if self._meta is None:
            res_gids = json.dumps([int(g) for g in self.sc_point.h5_gid_set])
            gid_counts = json.dumps(
                [float(np.round(n, 1)) for n in self.sc_point.gid_counts]
            )

            self._meta = pd.DataFrame(
                {
                    "gid": self.sc_point.gid,  # needed for collection
                    SupplyCurveField.LATITUDE: self.sc_point.latitude,
                    SupplyCurveField.LONGITUDE: self.sc_point.longitude,
                    SupplyCurveField.COUNTRY: self.sc_point.country,
                    SupplyCurveField.STATE: self.sc_point.state,
                    SupplyCurveField.COUNTY: self.sc_point.county,
                    SupplyCurveField.ELEVATION: self.sc_point.elevation,
                    SupplyCurveField.TIMEZONE: self.sc_point.timezone,
                    SupplyCurveField.SC_POINT_GID: self.sc_point.sc_point_gid,
                    SupplyCurveField.SC_ROW_IND: self.sc_point.sc_row_ind,
                    SupplyCurveField.SC_COL_IND: self.sc_point.sc_col_ind,
                    SupplyCurveField.RES_GIDS: res_gids,
                    SupplyCurveField.GID_COUNTS: gid_counts,
                    SupplyCurveField.N_GIDS: self.sc_point.n_gids,
                    SupplyCurveField.OFFSHORE: self.sc_point.offshore,
                    SupplyCurveField.AREA_SQ_KM: self.sc_point.area,
                },
                index=[self.sc_point.gid],
            )

        return self._meta

    @property
    def hub_height(self):
        """Get the integer SAM system config turbine hub height (meters)

        Returns
        -------
        int
        """
        return int(self.sam_sys_inputs["wind_turbine_hub_ht"])

    @property
    def res_df(self):
        """Get the reV compliant wind resource dataframe representing the
        aggregated and included wind resource in the current reV supply curve
        point at the turbine hub height. Includes a DatetimeIndex and columns
        for temperature, pressure, windspeed, and winddirection.

        Returns
        -------
        pd.DataFrame
        """
        if self._res_df is None:
            if self._pre_loaded_data is None:
                ti = self.sc_point.h5.time_index
            else:
                ti = self._pre_loaded_data.time_index

            wd = self.get_weighted_res_dir()
            ws = self.get_weighted_res_ts(f"windspeed_{self.hub_height}m")
            temp = self.get_weighted_res_ts(f"temperature_{self.hub_height}m")
            pres = self.get_weighted_res_ts(f"pressure_{self.hub_height}m")

            # convert mbar to atm
            if np.nanmax(pres) > 1000:
                pres *= 9.86923e-6

            data = {
                "temperature": temp,
                "pressure": pres,
                "windspeed": ws,
                "winddirection": wd,
            }

            if self.sam_sys_inputs.get("en_icing_cutoff"):
                rh = self.get_weighted_res_ts("relativehumidity_2m")
                data["relativehumidity"] = rh

            self._res_df = pd.DataFrame(data, index=ti)

            if "time_index_step" in self.original_sam_sys_inputs:
                ti_step = self.original_sam_sys_inputs["time_index_step"]
                self._res_df = self._res_df.iloc[::ti_step]

        return self._res_df

    @property
    def years(self):
        """Get the sorted list of analysis years.

        Returns
        -------
        list
        """
        return sorted(list(self.res_df.index.year.unique()))

    @property
    def annual_time_indexes(self):
        """Get an ordered list of single-year time index objects that matches
        the profile outputs from the wind_plant_ts object.

        Returns
        -------
        list
        """
        tis = []
        for year in self.years:
            ti = self.res_df.index[(self.res_df.index.year == year)]
            tis.append(WindPower.ensure_res_len(ti, ti))
        return tis

    @property
    def wind_dist(self):
        """Get the wind joint probability distribution and corresonding bin
        edges

        Returns
        -------
        wind_dist : np.ndarray
            2D array probability distribution of (windspeed, winddirection)
            normalized so the sum of all values = 1.
        ws_edges : np.ndarray
            1D array of windspeed (m/s) values that set the bin edges for the
            wind probability distribution. Same len as wind_dist.shape[0] + 1
        wd_edges : np.ndarray
            1D array of winddirections (deg) values that set the bin edges
            for the wind probability dist. Same len as wind_dist.shape[1] + 1
        """
        if self._wind_dist is None:
            ws_bins = JointPD._make_bins(*self._ws_bins)
            wd_bins = JointPD._make_bins(*self._wd_bins)

            hist_out = np.histogram2d(
                self.res_df["windspeed"],
                self.res_df["winddirection"],
                bins=(ws_bins, wd_bins),
            )
            self._wind_dist, self._ws_edges, self._wd_edges = hist_out
            self._wind_dist /= self._wind_dist.sum()

        return self._wind_dist, self._ws_edges, self._wd_edges

    def initialize_wind_plant_ts(self):
        """Initialize the annual wind plant timeseries analysis object(s) using
        the annual resource data and the sam system inputs from the optimized
        plant.

        Returns
        -------
        wind_plant_ts : dict
            Annual reV.SAM.generation.WindPower object(s) keyed by year.
        """
        wind_plant_ts = {}
        for year in self.years:
            res_df = self.res_df[(self.res_df.index.year == year)]
            sam_inputs = copy.deepcopy(self.sam_sys_inputs)

            if "lcoe_fcr" in self._out_req:
                lcoe_kwargs = self.get_lcoe_kwargs()
                sam_inputs.update(lcoe_kwargs)

            i_wp = WindPower(
                res_df, self.meta, sam_inputs, output_request=self._out_req
            )
            wind_plant_ts[year] = i_wp

        return wind_plant_ts

    @property
    def wind_plant_pd(self):
        """ReV WindPowerPD compute object for plant layout optimization based
        on wind joint probability distribution

        Returns
        -------
        reV.SAM.generation.WindPowerPD
        """

        if self._wind_plant_pd is None:
            wind_dist, ws_edges, wd_edges = self.wind_dist
            self._wind_plant_pd = WindPowerPD(
                ws_edges,
                wd_edges,
                wind_dist,
                self.meta,
                self.sam_sys_inputs,
                output_request=self._out_req,
            )
        return self._wind_plant_pd

    @property
    def wind_plant_ts(self):
        """ReV WindPower compute object(s) based on wind resource timeseries
        data keyed by year

        Returns
        -------
        dict
        """
        return self._wind_plant_ts

    @property
    def plant_optimizer(self):
        """Bespoke plant turbine placement optimizer object.

        Returns
        -------
        PlaceTurbines
        """
        if self._plant_optm is None:
            # put import here to delay breaking due to special dependencies
            from reV.bespoke.place_turbines import PlaceTurbines

            self._plant_optm = PlaceTurbines(
                self.wind_plant_pd,
                self.objective_function,
                self.capital_cost_function,
                self.fixed_operating_cost_function,
                self.variable_operating_cost_function,
                self.balance_of_system_cost_function,
                self.include_mask,
                self.pixel_side_length,
                self.min_spacing,
                self.wake_loss_multiplier,
            )

        return self._plant_optm

    def recalc_lcoe(self):
        """Recalculate the multi-year mean LCOE based on the multi-year mean
        annual energy production (AEP)"""

        if "lcoe_fcr-means" in self.outputs:
            lcoe_kwargs = self.get_lcoe_kwargs()

            logger.debug(
                "Recalulating multi-year mean LCOE using "
                "multi-year mean AEP."
            )

            fcr = lcoe_kwargs['fixed_charge_rate']
            cc = lcoe_kwargs['capital_cost']
            foc = lcoe_kwargs['fixed_operating_cost']
            voc = lcoe_kwargs['variable_operating_cost']
            aep = self.outputs['annual_energy-means']

            my_mean_lcoe = lcoe_fcr(fcr, cc, foc, aep, voc)

            self._outputs["lcoe_fcr-means"] = my_mean_lcoe
            self._meta[SupplyCurveField.MEAN_LCOE] = my_mean_lcoe

    def get_lcoe_kwargs(self):
        """Get a namespace of arguments for calculating LCOE based on the
        bespoke optimized wind plant capacity

        Returns
        -------
        lcoe_kwargs : dict
            kwargs for the SAM lcoe model. These are based on the original
            sam_sys_inputs, normalized to the original system_capacity, and
            updated based on the bespoke optimized system_capacity, includes
            fixed_charge_rate, system_capacity (kW), capital_cost ($),
            fixed_operating_cos ($), variable_operating_cost ($/kWh),
            balance_of_system_cost ($). Data source priority: outputs,
            plant_optimizer, original_sam_sys_inputs, meta
        """

        kwargs_map = {
            "fixed_charge_rate": SupplyCurveField.FIXED_CHARGE_RATE,
            "system_capacity": SupplyCurveField.CAPACITY_AC_MW,
            "capital_cost": SupplyCurveField.BESPOKE_CAPITAL_COST,
            "fixed_operating_cost": (
                SupplyCurveField.BESPOKE_FIXED_OPERATING_COST
            ),
            "variable_operating_cost": (
                SupplyCurveField.BESPOKE_VARIABLE_OPERATING_COST
            ),
            "balance_of_system_cost": (
                SupplyCurveField.BESPOKE_BALANCE_OF_SYSTEM_COST
            ),
        }
        lcoe_kwargs = {}

        for kwarg, meta_field in kwargs_map.items():
            if kwarg in self.outputs:
                lcoe_kwargs[kwarg] = self.outputs[kwarg]
            elif getattr(self.plant_optimizer, kwarg, None) is not None:
                lcoe_kwargs[kwarg] = getattr(self.plant_optimizer, kwarg)
            elif kwarg in self.original_sam_sys_inputs:
                lcoe_kwargs[kwarg] = self.original_sam_sys_inputs[kwarg]
            elif kwarg in self.meta:
                value = float(self.meta[kwarg].values[0])
                lcoe_kwargs[kwarg] = value
            elif meta_field in self.meta:
                value = float(self.meta[meta_field].values[0])
                if meta_field == SupplyCurveField.CAPACITY_AC_MW:
                    value *= 1000  # MW to kW
                lcoe_kwargs[kwarg] = value

        missing = [k for k in kwargs_map if k not in lcoe_kwargs]
        if any(missing):
            msg = (
                "Could not find these LCOE kwargs in outputs, "
                "plant_optimizer, original_sam_sys_inputs, or meta: {}".format(
                    missing
                )
            )
            logger.error(msg)
            raise KeyError(msg)

        bos = lcoe_kwargs.pop("balance_of_system_cost")
        lcoe_kwargs["capital_cost"] = lcoe_kwargs["capital_cost"] + bos
        return lcoe_kwargs

    @classmethod
    def check_dependencies(cls):
        """Check special dependencies for bespoke"""

        missing = []
        for name in cls.DEPENDENCIES:
            try:
                import_module(name)
            except ModuleNotFoundError:
                missing.append(name)

        if any(missing):
            msg = (
                "The reV bespoke module depends on the following special "
                "dependencies that were not found in the active "
                "environment: {}".format(missing)
            )
            logger.error(msg)
            raise ModuleNotFoundError(msg)

    @staticmethod
    def _check_sys_inputs(plant1, plant2,
                          ignore=('wind_resource_model_choice',
                                  'wind_resource_data',
                                  'wind_turbine_powercurve_powerout',
                                  'hourly',
                                  'capital_cost',
                                  'fixed_operating_cost',
                                  'variable_operating_cost',
                                  'balance_of_system_cost',
                                  'base_capital_cost',
                                  'base_fixed_operating_cost',
                                  'base_variable_operating_cost')):
        """Check two reV-SAM models for matching system inputs.

        Parameters
        ----------
        plant1/plant2 : reV.SAM.generation.WindPower
            Two WindPower analysis objects to check.
        """
        bad = []
        for k, v in plant1.sam_sys_inputs.items():
            if k not in plant2.sam_sys_inputs or str(v) != str(
                plant2.sam_sys_inputs[k]
            ):
                bad.append(k)
        bad = [b for b in bad if b not in ignore]
        if any(bad):
            msg = "Inputs no longer match: {}".format(bad)
            logger.error(msg)
            raise RuntimeError(msg)

    def run_wind_plant_ts(self):
        """Run the wind plant multi-year timeseries analysis and export output
        requests to outputs property.

        Returns
        -------
        outputs : dict
            Output dictionary for the full BespokeSinglePlant object. The
            multi-year timeseries data is also exported to the
            BespokeSinglePlant.outputs property.
        """

        logger.debug(
            "Running {} years of SAM timeseries analysis for {}".format(
                len(self.years), self
            )
        )
        self._wind_plant_ts = self.initialize_wind_plant_ts()
        for year, plant in self.wind_plant_ts.items():
            self._check_sys_inputs(plant, self.wind_plant_pd)
            try:
                plant.run_gen_and_econ()
            except Exception as e:
                msg = (
                    "{} failed while trying to run SAM WindPower "
                    "timeseries analysis for {}".format(self, year)
                )
                logger.exception(msg)
                raise RuntimeError(msg) from e

            for k, v in plant.outputs.items():
                self._outputs[k + "-{}".format(year)] = v

        means = {}
        for k1, v1 in self._outputs.items():
            if isinstance(v1, Number) and parse_year(k1, option="boolean"):
                year = parse_year(k1)
                base_str = k1.replace(str(year), "")
                all_values = [
                    v2 for k2, v2 in self._outputs.items() if base_str in k2
                ]
                means[base_str + "means"] = np.mean(all_values)

        self._outputs.update(means)

        self._meta[SupplyCurveField.MEAN_RES] = self.res_df["windspeed"].mean()
        self._meta[SupplyCurveField.MEAN_CF_DC] = np.nan
        self._meta[SupplyCurveField.MEAN_CF_AC] = np.nan
        self._meta[SupplyCurveField.MEAN_LCOE] = np.nan
        self._meta[SupplyCurveField.SC_POINT_ANNUAL_ENERGY_MW] = np.nan
        # copy dataset outputs to meta data for supply curve table summary
        if "cf_mean-means" in self.outputs:
            self._meta.loc[:, SupplyCurveField.MEAN_CF_AC] = self.outputs[
                "cf_mean-means"
            ]
        if "lcoe_fcr-means" in self.outputs:
            self._meta.loc[:, SupplyCurveField.MEAN_LCOE] = self.outputs[
                "lcoe_fcr-means"
            ]
            self.recalc_lcoe()
        if "annual_energy-means" in self.outputs:
            self._meta[SupplyCurveField.SC_POINT_ANNUAL_ENERGY_MW] = (
                self.outputs["annual_energy-means"] / 1000
            )

        logger.debug("Timeseries analysis complete!")

        return self.outputs

    def run_plant_optimization(self):
        """Run the wind plant layout optimization and export outputs
        to outputs property.

        Returns
        -------
        outputs : dict
            Output dictionary for the full BespokeSinglePlant object. The
            layout optimization output data is also exported to the
            BespokeSinglePlant.outputs property.
        """

        logger.debug("Running plant layout optimization for {}".format(self))
        try:
            self.plant_optimizer.place_turbines(**self.ga_kwargs)
        except Exception as e:
            msg = (
                "{} failed while trying to run the "
                "turbine placement optimizer".format(self)
            )
            logger.exception(msg)
            raise RuntimeError(msg) from e

        self._outputs["full_polygons"] = self.plant_optimizer.full_polygons
        self._outputs["packing_polygons"] = (
            self.plant_optimizer.packing_polygons
        )
        system_capacity_kw = self.plant_optimizer.capacity
        self._outputs["system_capacity"] = system_capacity_kw

        txc = [int(np.round(c)) for c in self.plant_optimizer.turbine_x]
        tyc = [int(np.round(c)) for c in self.plant_optimizer.turbine_y]
        pxc = [int(np.round(c)) for c in self.plant_optimizer.x_locations]
        pyc = [int(np.round(c)) for c in self.plant_optimizer.y_locations]

        txc = json.dumps(txc)
        tyc = json.dumps(tyc)
        pxc = json.dumps(pxc)
        pyc = json.dumps(pyc)

        self._meta[SupplyCurveField.TURBINE_X_COORDS] = txc
        self._meta[SupplyCurveField.TURBINE_Y_COORDS] = tyc
        self._meta[SupplyCurveField.POSSIBLE_X_COORDS] = pxc
        self._meta[SupplyCurveField.POSSIBLE_Y_COORDS] = pyc

        self._meta[SupplyCurveField.N_TURBINES] = self.plant_optimizer.nturbs
        self._meta["avg_sl_dist_to_center_m"] = (
            self.plant_optimizer.avg_sl_dist_to_center_m
        )
        self._meta["avg_sl_dist_to_medoid_m"] = (
            self.plant_optimizer.avg_sl_dist_to_medoid_m
        )
        self._meta["nn_conn_dist_m"] = self.plant_optimizer.nn_conn_dist_m
        self._meta[SupplyCurveField.BESPOKE_AEP] = self.plant_optimizer.aep
        self._meta[SupplyCurveField.BESPOKE_OBJECTIVE] = (
            self.plant_optimizer.objective
        )
        self._meta[SupplyCurveField.BESPOKE_CAPITAL_COST] = (
            self.plant_optimizer.capital_cost
        )
        self._meta[SupplyCurveField.BESPOKE_FIXED_OPERATING_COST] = (
            self.plant_optimizer.fixed_operating_cost
        )
        self._meta[SupplyCurveField.BESPOKE_VARIABLE_OPERATING_COST] = (
            self.plant_optimizer.variable_operating_cost
        )
        self._meta[SupplyCurveField.BESPOKE_BALANCE_OF_SYSTEM_COST] = (
            self.plant_optimizer.balance_of_system_cost
        )
        self._meta[SupplyCurveField.INCLUDED_AREA] = self.plant_optimizer.area
        self._meta[SupplyCurveField.INCLUDED_AREA_CAPACITY_DENSITY] = (
            self.plant_optimizer.capacity_density
        )
        self._meta[SupplyCurveField.CONVEX_HULL_AREA] = (
            self.plant_optimizer.convex_hull_area
        )
        self._meta[SupplyCurveField.CONVEX_HULL_CAPACITY_DENSITY] = (
            self.plant_optimizer.convex_hull_capacity_density
        )
        self._meta[SupplyCurveField.FULL_CELL_CAPACITY_DENSITY] = (
            self.plant_optimizer.full_cell_capacity_density
        )

        # copy dataset outputs to meta data for supply curve table summary
        # convert SAM system capacity in kW to reV supply curve cap in MW
        capacity_ac_mw = system_capacity_kw / 1e3
        self._meta[SupplyCurveField.CAPACITY_AC_MW] = capacity_ac_mw
        self._meta[SupplyCurveField.CAPACITY_DC_MW] = np.nan

        # add required ReEDS multipliers to meta
        baseline_cost = self.plant_optimizer.capital_cost_per_kw(
            capacity_mw=self._baseline_cap_mw
        )
        eos_mult = (self.plant_optimizer.capital_cost
                    / self.plant_optimizer.capacity
                    / baseline_cost)
        reg_mult_cc = self.sam_sys_inputs.get(
            "capital_cost_multiplier", 1)
        reg_mult_foc = self.sam_sys_inputs.get(
            "fixed_operating_cost_multiplier", 1)
        reg_mult_voc = self.sam_sys_inputs.get(
            "variable_operating_cost_multiplier", 1)
        reg_mult_bos = self.sam_sys_inputs.get(
            "balance_of_system_cost_multiplier", 1)

        self._meta[SupplyCurveField.EOS_MULT] = eos_mult
        self._meta[SupplyCurveField.REG_MULT] = reg_mult_cc

        self._meta[SupplyCurveField.COST_SITE_OCC_USD_PER_AC_MW] = (
            (self.plant_optimizer.capital_cost
             + self.plant_optimizer.balance_of_system_cost)
            / capacity_ac_mw
        )
        self._meta[SupplyCurveField.COST_BASE_OCC_USD_PER_AC_MW] = (
            (self.plant_optimizer.capital_cost / eos_mult / reg_mult_cc
             + self.plant_optimizer.balance_of_system_cost / reg_mult_bos)
            / capacity_ac_mw
        )
        self._meta[SupplyCurveField.COST_SITE_FOC_USD_PER_AC_MW] = (
            self.plant_optimizer.fixed_operating_cost
            / capacity_ac_mw
        )
        self._meta[SupplyCurveField.COST_BASE_FOC_USD_PER_AC_MW] = (
            self.plant_optimizer.fixed_operating_cost
            / reg_mult_foc
            / capacity_ac_mw
        )
        self._meta[SupplyCurveField.COST_SITE_VOC_USD_PER_AC_MW] = (
            self.plant_optimizer.variable_operating_cost
            / capacity_ac_mw
        )
        self._meta[SupplyCurveField.COST_BASE_VOC_USD_PER_AC_MW] = (
            self.plant_optimizer.variable_operating_cost
            / reg_mult_voc
            / capacity_ac_mw
        )
        self._meta[SupplyCurveField.FIXED_CHARGE_RATE] = (
            self.plant_optimizer.fixed_charge_rate
        )

        logger.debug("Plant layout optimization complete!")
        return self.outputs

    def agg_data_layers(self):
        """Aggregate optional data layers if requested and save to self.meta"""
        if self._data_layers is not None:
            logger.debug(
                "Aggregating {} extra data layers.".format(
                    len(self._data_layers)
                )
            )
            point_summary = self.meta.to_dict()
            point_summary = self.sc_point.agg_data_layers(
                point_summary, self._data_layers
            )
            self._meta = pd.DataFrame(point_summary)
            logger.debug("Finished aggregating extra data layers.")

    @property
    def outputs(self):
        """Saved outputs for the single wind plant bespoke optimization.

        Returns
        -------
        dict
        """
        return self._outputs

    @classmethod
    def run(cls, *args, **kwargs):
        """Run the bespoke optimization for a single wind plant.

        Parameters
        ----------
        See the class initialization parameters.

        Returns
        -------
        bsp : dict
            Bespoke single plant outputs namespace keyed by dataset name
            including a dataset "meta" for the BespokeSinglePlant meta data.
        """

        with cls(*args, **kwargs) as bsp:
            if bsp._prior_meta:
                logger.debug(
                    "Skipping bespoke plant optimization for gid {}. "
                    "Received prior meta data for this point.".format(bsp.gid)
                )
            else:
                _ = bsp.run_plant_optimization()

            _ = bsp.run_wind_plant_ts()
            bsp.agg_data_layers()

            meta = bsp.meta
            out = bsp.outputs
            out["meta"] = meta
            for year, ti in zip(bsp.years, bsp.annual_time_indexes):
                out["time_index-{}".format(year)] = ti

        return out


class BespokeWindPlants(BaseAggregation):
    """BespokeWindPlants"""

    def __init__(self, excl_fpath, res_fpath, tm_dset, objective_function,
                 capital_cost_function, fixed_operating_cost_function,
                 variable_operating_cost_function,
                 balance_of_system_cost_function, project_points,
                 sam_files, min_spacing='5x', wake_loss_multiplier=1,
                 ga_kwargs=None, output_request=('system_capacity', 'cf_mean'),
                 ws_bins=(0.0, 20.0, 5.0), wd_bins=(0.0, 360.0, 45.0),
                 excl_dict=None, area_filter_kernel='queen', min_area=None,
                 resolution=64, excl_area=None, data_layers=None,
                 pre_extract_inclusions=False, eos_mult_baseline_cap_mw=200,
                 prior_run=None, gid_map=None, bias_correct=None,
                 pre_load_data=False):
        """reV bespoke analysis class.

        Much like generation, ``reV`` bespoke analysis runs SAM
        simulations by piping in renewable energy resource data (usually
        from the WTK), loading the SAM config, and then executing the
        :py:class:`PySAM.Windpower.Windpower` compute module.
        However, unlike ``reV`` generation, bespoke analysis is
        performed on the supply-curve grid resolution, and the plant
        layout is optimized for every supply-curve point based on an
        optimization objective specified by the user. See the NREL
        publication on the bespoke methodology for more information.

        See the documentation for the ``reV`` SAM class (e.g.
        :class:`reV.SAM.generation.WindPower`,
        :class:`reV.SAM.generation.PvWattsv8`,
        :class:`reV.SAM.generation.Geothermal`, etc.) for info on the
        allowed and/or required SAM config file inputs.

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions data HDF5 file. The exclusions HDF5
            file should contain the layers specified in `excl_dict`
            and `data_layers`. These layers may also be spread out
            across multiple HDF5 files, in which case this input should
            be a list or tuple of filepaths pointing to the files
            containing the layers. Note that each data layer must be
            uniquely defined (i.e.only appear once and in a single
            input file).
        res_fpath : str
            Unix shell style path to wind resource HDF5 file in NREL WTK
            format. Can also be a path including a wildcard input like
            ``/h5_dir/prefix*suffix`` to run bespoke on multiple years
            of resource data. Can also be an explicit list of resource
            HDF5 file paths, which themselves can contain wildcards. If
            multiple files are specified in this way, they must have the
            same coordinates but can have different time indices (i.e.
            different years). This input must be readable by
            :py:class:`rex.multi_year_resource.MultiYearWindResource`
            (i.e. the resource data conform to the
            `rex data format <https://tinyurl.com/3fy7v5kx>`_). This
            means the data file(s) must contain a 1D ``time_index``
            dataset indicating the UTC time of observation, a 1D
            ``meta`` dataset represented by a DataFrame with
            site-specific columns, and 2D resource datasets that match
            the dimensions of (time_index, meta). The time index must
            start at 00:00 of January 1st of the year under
            consideration, and its shape must be a multiple of 8760.
        tm_dset : str
            Dataset name in the `excl_fpath` file containing the
            techmap (exclusions-to-resource mapping data). This data
            layer links the supply curve GID's to the generation GID's
            that are used to evaluate the performance metrics of each
            wind plant. By default, the generation GID's are assumed to
            match the resource GID's, but this mapping can be customized
            via the `gid_map` input (see the documentation for `gid_map`
            for more details).

            .. Important:: This dataset uniquely couples the (typically
              high-resolution) exclusion layers to the (typically
              lower-resolution) resource data. Therefore, a separate
              techmap must be used for every unique combination of
              resource and exclusion coordinates.

        objective_function : str
            The objective function of the optimization written out as a
            string. This expression should compute the objective to be
            minimized during layout optimization. Variables available
            for computation are:

                - ``n_turbines``: the number of turbines
                - ``system_capacity``: wind plant capacity
                - ``aep``: annual energy production
                - ``avg_sl_dist_to_center_m``: Average straight-line
                  distance to the supply curve point center from all
                  turbine locations (in m). Useful for computing plant
                  BOS costs.
                - ``avg_sl_dist_to_medoid_m``: Average straight-line
                  distance to the medoid of all turbine locations
                  (in m). Useful for computing plant BOS costs.
                - ``nn_conn_dist_m``: Total BOS connection distance
                  using nearest-neighbor connections. This variable is
                  only available for the
                  ``balance_of_system_cost_function`` equation.
                - ``fixed_charge_rate``: user input fixed_charge_rate if
                  included as part of the sam system config.
                - ``capital_cost``: plant capital cost as evaluated
                  by `capital_cost_function`
                - ``fixed_operating_cost``: plant fixed annual operating
                  cost as evaluated by `fixed_operating_cost_function`
                - ``variable_operating_cost``: plant variable annual
                  operating cost as evaluated by
                  `variable_operating_cost_function`
                - ``balance_of_system_cost``: plant balance of system
                  cost as evaluated by `balance_of_system_cost_function`
                - ``self.wind_plant``: the SAM wind plant object,
                  through which all SAM variables can be accessed

        capital_cost_function : str
            The plant capital cost function written out as a string.
            This expression must return the total plant capital cost in
            $. This expression has access to the same variables as the
            `objective_function` argument above.
        fixed_operating_cost_function : str
            The plant annual fixed operating cost function written out
            as a string. This expression must return the fixed operating
            cost in $/year. This expression has access to the same
            variables as the `objective_function` argument above.
        variable_operating_cost_function : str
            The plant annual variable operating cost function written
            out as a string. This expression must return the variable
            operating cost in $/kWh. This expression has access to the
            same variables as the `objective_function` argument above.
            You can set this to "0" to effectively ignore variable
            operating costs.
        balance_of_system_cost_function : str
            The plant balance-of-system cost function as a string, must
            return the variable operating cost in $. Has access to the
            same variables as the objective_function. You can set this
            to "0" to effectively ignore balance-of-system costs.
        project_points : int | list | tuple | str | dict | pd.DataFrame | slice
            Input specifying which sites to process. A single integer
            representing the supply curve GID of a site may be specified
            to evaluate ``reV`` at a supply curve point. A list or tuple
            of integers (or slice) representing the supply curve GIDs of
            multiple sites can be specified to evaluate ``reV`` at
            multiple specific locations. A string pointing to a project
            points CSV file may also be specified. Typically, the CSV
            contains the following columns:

                - ``gid``: Integer specifying the supply curve GID of
                  each site.
                - ``config``: Key in the `sam_files` input dictionary
                  (see below) corresponding to the SAM configuration to
                  use for each particular site. This value can also be
                  ``None`` (or left out completely) if you specify only
                  a single SAM configuration file as the `sam_files`
                  input.

            The CSV file may also contain site-specific inputs by
            including a column named after a config keyword (e.g. a
            column called ``capital_cost`` may be included to specify a
            site-specific capital cost value for each location). Columns
            that do not correspond to a config key may also be included,
            but they will be ignored. The CSV file input can also have
            these extra, optional columns:

                - ``capital_cost_multiplier``
                - ``fixed_operating_cost_multiplier``
                - ``variable_operating_cost_multiplier``
                - ``balance_of_system_cost_multiplier``

            These particular inputs are treated as multipliers to be
            applied to the respective cost curves
            (`capital_cost_function`, `fixed_operating_cost_function`,
            `variable_operating_cost_function`, and
            `balance_of_system_cost_function`) both during and
            after the optimization. A DataFrame following the same
            guidelines as the CSV input (or a dictionary that can be
            used to initialize such a DataFrame) may be used for this
            input as well. If you would like to obtain all available
            ``reV`` supply curve points to run, you can use the
            :class:`reV.supply_curve.extent.SupplyCurveExtent` class
            like so::

                import pandas as pd
                from reV.supply_curve.extent import SupplyCurveExtent

                excl_fpath = "..."
                resolution = ...
                with SupplyCurveExtent(excl_fpath, resolution) as sc:
                    points = sc.valid_sc_points(tm_dset).tolist()
                    points = pd.DataFrame({"gid": points})
                    points["config"] = "default"  # or a list of config choices

                # Use the points directly or save them to csv for CLI usage
                points.to_csv("project_points.csv", index=False)

        sam_files : dict | str
            A dictionary mapping SAM input configuration ID(s) to SAM
            configuration(s). Keys are the SAM config ID(s) which
            correspond to the ``config`` column in the project points
            CSV. Values for each key are either a path to a
            corresponding SAM config file or a full dictionary
            of SAM config inputs. For example::

                sam_files = {
                    "default": "/path/to/default/sam.json",
                    "onshore": "/path/to/onshore/sam_config.yaml",
                    "offshore": {
                        "sam_key_1": "sam_value_1",
                        "sam_key_2": "sam_value_2",
                        ...
                    },
                    ...
                }

            This input can also be a string pointing to a single SAM
            config file. In this case, the ``config`` column of the
            CSV points input should be set to ``None`` or left out
            completely. See the documentation for the ``reV`` SAM class
            (e.g. :class:`reV.SAM.generation.WindPower`,
            :class:`reV.SAM.generation.PvWattsv8`,
            :class:`reV.SAM.generation.Geothermal`, etc.) for
            info on the allowed and/or required SAM config file inputs.
        min_spacing : float | int | str, optional
            Minimum spacing between turbines (in meters). This input can
            also be a string like "5x", which is interpreted as 5 times
            the turbine rotor diameter. By default, ``"5x"``.
        wake_loss_multiplier : float, optional
            A multiplier used to scale the annual energy lost due to
            wake losses.

            .. WARNING:: This multiplier will ONLY be applied during the
               optimization process and will NOT come through in output
               values such as the hourly profiles, aep, any of the cost
               functions, or even the output objective.

            By default, ``1``.
        ga_kwargs : dict, optional
            Dictionary of keyword arguments to pass to GA
            initialization. If ``None``, default initialization values
            are used. See
            :class:`~reV.bespoke.gradient_free.GeneticAlgorithm` for
            a description of the allowed keyword arguments.
            By default, ``None``.
        output_request : list | tuple, optional
            Outputs requested from the SAM windpower simulation after
            the bespoke plant layout optimization. Can be any of the
            parameters in the "Outputs" group of the PySAM module
            :py:class:`PySAM.Windpower.Windpower.Outputs`, PySAM module.
            This list can also include a select number of SAM
            config/resource parameters to include in the output:
            any key in any of the
            `output attribute JSON files <https://tinyurl.com/4bmrpe3j/>`_
            may be requested. Time-series profiles requested via this
            input are output in UTC. This input can also be used to
            request resource means like ``"ws_mean"``,
            ``"windspeed_mean"``, ``"temperature_mean"``, and
            ``"pressure_mean"``. By default,
            ``('system_capacity', 'cf_mean')``.
        ws_bins : tuple, optional
            A 3-entry tuple with ``(start, stop, step)`` for the
            windspeed binning of the wind joint probability
            distribution. The stop value is inclusive, so
            ``ws_bins=(0, 20, 5)`` would result in four bins with bin
            edges (0, 5, 10, 15, 20). By default, ``(0.0, 20.0, 5.0)``.
        wd_bins : tuple, optional
            A 3-entry tuple with ``(start, stop, step)`` for the wind
            direction binning of the wind joint probability
            distribution. The stop value is inclusive, so
            ``wd_bins=(0, 360, 90)`` would result in four bins with bin
            edges (0, 90, 180, 270, 360).
            By default, ``(0.0, 360.0, 45.0)``.
        excl_dict : dict, optional
            Dictionary of exclusion keyword arguments of the format
            ``{layer_dset_name: {kwarg: value}}``, where
            ``layer_dset_name`` is a dataset in the exclusion h5 file
            and the ``kwarg: value`` pair is a keyword argument to
            the :class:`reV.supply_curve.exclusions.LayerMask` class.
            For example::

                excl_dict = {
                    "typical_exclusion": {
                        "exclude_values": 255,
                    },
                    "another_exclusion": {
                        "exclude_values": [2, 3],
                        "weight": 0.5
                    },
                    "exclusion_with_nodata": {
                        "exclude_range": [10, 100],
                        "exclude_nodata": True,
                        "nodata_value": -1
                    },
                    "partial_setback": {
                        "use_as_weights": True
                    },
                    "height_limit": {
                        "exclude_range": [0, 200]
                    },
                    "slope": {
                        "include_range": [0, 20]
                    },
                    "developable_land": {
                        "force_include_values": 42
                    },
                    "more_developable_land": {
                        "force_include_range": [5, 10]
                    },
                    ...
                }

            Note that all the keys given in this dictionary should be
            datasets of the `excl_fpath` file. If ``None`` or empty
            dictionary, no exclusions are applied. By default, ``None``.
        area_filter_kernel : {"queen", "rook"}, optional
            Contiguous area filter method to use on final exclusions
            mask. The filters are defined as::

                # Queen:     # Rook:
                [[1,1,1],    [[0,1,0],
                 [1,1,1],     [1,1,1],
                 [1,1,1]]     [0,1,0]]

            These filters define how neighboring pixels are "connected".
            Once pixels in the final exclusion layer are connected, the
            area of each resulting cluster is computed and compared
            against the `min_area` input. Any cluster with an area
            less than `min_area` is excluded from the final mask.
            This argument has no effect if `min_area` is ``None``.
            By default, ``"queen"``.
        min_area : float, optional
            Minimum area (in km\ :sup:`2`) required to keep an isolated
            cluster of (included) land within the resulting exclusions
            mask. Any clusters of land with areas less than this value
            will be marked as exclusions. See the documentation for
            `area_filter_kernel` for an explanation of how the area of
            each land cluster is computed. If ``None``, no area
            filtering is performed. By default, ``None``.
        resolution : int, optional
            Supply Curve resolution. This value defines how many pixels
            are in a single side of a supply curve cell. For example,
            a value of ``64`` would generate a supply curve where the
            side of each supply curve cell is ``64x64`` exclusion
            pixels. By default, ``64``.
        excl_area : float, optional
            Area of a single exclusion mask pixel (in km\ :sup:`2`).
            If ``None``, this value will be inferred from the profile
            transform attribute in `excl_fpath`. By default, ``None``.
        data_layers : dict, optional
            Dictionary of aggregation data layers of the format::

                data_layers = {
                    "output_layer_name": {
                        "dset": "layer_name",
                        "method": "mean",
                        "fpath": "/path/to/data.h5"
                    },
                    "another_output_layer_name": {
                        "dset": "input_layer_name",
                        "method": "mode",
                        # optional "fpath" key omitted
                    },
                    ...
                }

            The ``"output_layer_name"`` is the column name under which
            the aggregated data will appear in the meta DataFrame of the
            output file. The ``"output_layer_name"`` does not have to
            match the ``dset`` input value. The latter should match
            the layer name in the HDF5 from which the data to aggregate
            should be pulled. The ``method`` should be one of
            ``{"mode", "mean", "min", "max", "sum", "category"}``,
            describing how the high-resolution data should be aggregated
            for each supply curve point. ``fpath`` is an optional key
            that can point to an HDF5 file containing the layer data. If
            left out, the data is assumed to exist in the file(s)
            specified by the `excl_fpath` input. If ``None``, no data
            layer aggregation is performed. By default, ``None``.
        pre_extract_inclusions : bool, optional
            Optional flag to pre-extract/compute the inclusion mask from
            the `excl_dict` input. It is typically faster to compute
            the inclusion mask on the fly with parallel workers.
            By default, ``False``.
        eos_mult_baseline_cap_mw : int | float, optional
            Baseline plant capacity (MW) used to calculate economies of
            scale (EOS) multiplier from the `capital_cost_function`. EOS
            multiplier is calculated as the $-per-kW of the wind plant
            divided by the $-per-kW of a plant with this baseline
            capacity. By default, `200` (MW), which aligns the baseline
            with ATB assumptions. See here: https://tinyurl.com/y85hnu6h.
        prior_run : str, optional
            Optional filepath to a bespoke output HDF5 file belonging to
            a prior run. If specified, this module will only run the
            timeseries power generation step and assume that all of the
            wind plant layouts are fixed from the prior run. The meta
            data of this file must contain the following columns
            (automatically satisfied if the HDF5 file was generated by
            ``reV`` bespoke):

                - ``capacity`` : Capacity of the plant, in MW.
                - ``turbine_x_coords``: A string representation of a
                  python list containing the X coordinates (in m; origin
                  of cell at bottom left) of the turbines within the
                  plant (supply curve cell).
                - ``turbine_y_coords`` : A string representation of a
                  python list containing the Y coordinates (in m; origin
                  of cell at bottom left) of the turbines within the
                  plant (supply curve cell).

            If ``None``, no previous run data is considered.
            By default, ``None``
        gid_map : str | dict, optional
            Mapping of unique integer generation gids (keys) to single
            integer resource gids (values). This enables unique
            generation gids in the project points to map to non-unique
            resource gids, which can be useful when evaluating multiple
            resource datasets in ``reV`` (e.g., forecasted ECMWF
            resource data to complement historical WTK meteorology).
            This input can be a pre-extracted dictionary or a path to a
            JSON or CSV file. If this input points to a CSV file, the
            file must have the columns ``gid`` (which matches the
            project points) and ``gid_map`` (gids to extract from the
            resource input). If ``None``, the GID values in the project
            points are assumed to match the resource GID values.
            By default, ``None``.
        bias_correct : str | pd.DataFrame, optional
            Optional DataFrame or CSV filepath to a wind or solar
            resource bias correction table. This has columns:

                - ``gid``: GID of site (can be index name of dataframe)
                - ``method``: function name from ``rex.bias_correction`` module

            The ``gid`` field should match the true resource ``gid`` regardless
            of the optional ``gid_map`` input. Only ``windspeed`` **or**
            ``GHI`` + ``DNI`` + ``DHI`` are corrected, depending on the
            technology (wind for the former, PV or CSP for the latter). See the
            functions in the ``rex.bias_correction`` module for available
            inputs for ``method``. Any additional kwargs required for the
            requested ``method`` can be input as additional columns in the
            ``bias_correct`` table e.g., for linear bias correction functions
            you can include ``scalar`` and ``adder`` inputs as columns in the
            ``bias_correct`` table on a site-by-site basis. If ``None``, no
            corrections are applied. By default, ``None``.
        pre_load_data : bool, optional
            Option to pre-load resource data. This step can be
            time-consuming up front, but it drastically reduces the
            number of parallel reads to the `res_fpath` HDF5 file(s),
            and can have a significant overall speedup on systems with
            slow parallel I/O capabilities. Pre-loaded data can use a
            significant amount of RAM, so be sure to split execution
            across many nodes (e.g. 100 nodes, 36 workers each for
            CONUS) or request large amounts of memory for a smaller
            number of nodes. By default, ``False``.
        """

        log_versions(logger)
        logger.info('Initializing BespokeWindPlants...')
        logger.info('Resource filepath: {}'.format(res_fpath))
        logger.info('Exclusion filepath: {}'.format(excl_fpath))
        logger.debug('Exclusion dict: {}'.format(excl_dict))
        logger.info('Bespoke objective function: {}'
                    .format(objective_function))
        logger.info('Bespoke capital cost function: {}'
                    .format(capital_cost_function))
        logger.info('Bespoke fixed operating cost function: {}'
                    .format(fixed_operating_cost_function))
        logger.info('Bespoke variable operating cost function: {}'
                    .format(variable_operating_cost_function))
        logger.info('Bespoke balance of system cost function: {}'
                    .format(balance_of_system_cost_function))
        logger.info('Bespoke wake loss multiplier: {}'
                    .format(wake_loss_multiplier))
        logger.info('Bespoke GA initialization kwargs: {}'.format(ga_kwargs))

        logger.info(
            "Bespoke pre-extracting exclusions: {}".format(
                pre_extract_inclusions
            )
        )
        logger.info(
            "Bespoke pre-extracting resource data: {}".format(pre_load_data)
        )
        logger.info("Bespoke prior run: {}".format(prior_run))
        logger.info("Bespoke GID map: {}".format(gid_map))
        logger.info("Bespoke bias correction table: {}".format(bias_correct))

        BespokeSinglePlant.check_dependencies()

        self._project_points = self._parse_points(project_points, sam_files)

        super().__init__(
            excl_fpath,
            tm_dset,
            excl_dict=excl_dict,
            area_filter_kernel=area_filter_kernel,
            min_area=min_area,
            resolution=resolution,
            excl_area=excl_area,
            gids=self._project_points.gids,
            pre_extract_inclusions=pre_extract_inclusions,
        )

        self._res_fpath = res_fpath
        self._obj_fun = objective_function
        self._cap_cost_fun = capital_cost_function
        self._foc_fun = fixed_operating_cost_function
        self._voc_fun = variable_operating_cost_function
        self._bos_fun = balance_of_system_cost_function
        self._min_spacing = min_spacing
        self._wake_loss_multiplier = wake_loss_multiplier
        self._ga_kwargs = ga_kwargs or {}
        self._output_request = SAMOutputRequest(output_request)
        self._ws_bins = ws_bins
        self._wd_bins = wd_bins
        self._data_layers = data_layers
        self._eos_mult_baseline_cap_mw = eos_mult_baseline_cap_mw
        self._prior_meta = self._parse_prior_run(prior_run)
        self._gid_map = BespokeSinglePlant._parse_gid_map(gid_map)
        self._bias_correct = Gen._parse_bc(bias_correct)
        self._outputs = {}
        self._check_files()

        self._pre_loaded_data = None
        self._pre_load_data(pre_load_data)

        self._slice_lookup = None

        logger.info(
            "Initialized BespokeWindPlants with project points: {}".format(
                self._project_points
            )
        )

    @staticmethod
    def _parse_points(points, sam_configs):
        """Parse a project points object using a project points file

        Parameters
        ----------
        points : int | slice | list | str | PointsControl | None
            Slice or list specifying project points, string pointing to a
            project points csv, or a fully instantiated PointsControl object.
            Can also be a single site integer value. Points csv should have
            `SiteDataField.GID` and 'config' column, the config maps to the
            sam_configs dict keys.
        sam_configs : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s). Keys are the SAM
            config ID(s) which map to the config column in the project points
            CSV. Values are either a JSON SAM config file or dictionary of SAM
            config inputs. Can also be a single config file path or a
            pre loaded SAMConfig object.

        Returns
        -------
        ProjectPoints : ~reV.config.project_points.ProjectPoints
            Project points object laying out the supply curve gids to
            analyze.
        """
        pc = Gen.get_pc(
            points,
            points_range=None,
            sam_configs=sam_configs,
            tech="windpower",
            sites_per_worker=1,
        )

        return pc.project_points

    @staticmethod
    def _parse_prior_run(prior_run):
        """Extract bespoke meta data from prior run and verify that the run is
        compatible with the new job specs.

        Parameters
        ----------
        prior_run : str | None
            Optional filepath to a bespoke output .h5 file belonging to a prior
            run. This will only run the timeseries power generation step and
            assume that all of the wind plant layouts are fixed given the prior
            run. The meta data of this file needs columns "capacity",
            "turbine_x_coords", and "turbine_y_coords".

        Returns
        -------
        meta : pd.DataFrame | None
            Meta data from the previous bespoke run. This includes the
            previously optimized wind farm layouts. All of the nested list
            columns will be json loaded.
        """

        meta = None

        if prior_run is not None:
            assert os.path.isfile(prior_run)
            assert prior_run.endswith(".h5")

            with Outputs(prior_run, mode="r") as f:
                meta = f.meta
                meta = meta.rename(columns=SupplyCurveField.map_from_legacy())

            # pylint: disable=no-member
            for col in meta.columns:
                val = meta[col].values[0]
                if isinstance(val, str) and val[0] == "[" and val[-1] == "]":
                    meta[col] = meta[col].apply(json.loads)

        return meta

    def _get_prior_meta(self, gid):
        """Get the meta data for a given gid from the prior run (if available)

        Parameters
        ----------
        gid : int
            SC point gid for site to pull prior meta for.

        Returns
        -------
        meta : pd.DataFrame
            Prior meta data for just the requested gid.
        """
        meta = None

        if self._prior_meta is not None:
            mask = self._prior_meta[SupplyCurveField.SC_POINT_GID] == gid
            if any(mask):
                meta = self._prior_meta[mask]

        return meta

    def _check_files(self):
        """Do a preflight check on input files"""

        paths = self._excl_fpath
        if isinstance(self._excl_fpath, str):
            paths = [self._excl_fpath]

        for path in paths:
            if not os.path.exists(path):
                raise FileNotFoundError(
                    "Could not find required exclusions file: " "{}".format(
                        path
                    )
                )

        with ExclusionLayers(paths) as excl:
            if self._tm_dset not in excl:
                raise FileInputError(
                    'Could not find techmap dataset "{}" '
                    "in the exclusions file(s): {}".format(
                        self._tm_dset, paths
                    )
                )

        # just check that this file exists, cannot check res_fpath if *glob
        with MultiYearWindResource(self._res_fpath) as f:
            assert any(f.dsets)

    def _pre_load_data(self, pre_load_data):
        """Pre-load resource data, if requested."""
        if not pre_load_data:
            return

        sc_gid_to_hh = {
            gid: self._hh_for_sc_gid(gid)
            for gid in self._project_points.df[ResourceMetaField.GID]
        }

        with ExclusionLayers(self._excl_fpath) as excl:
            tm = excl[self._tm_dset]

        scp_kwargs = {"shape": self.shape, "resolution": self._resolution}
        slices = {
            gid: SupplyCurvePoint.get_agg_slices(gid=gid, **scp_kwargs)
            for gid in self._project_points.df[ResourceMetaField.GID]
        }

        sc_gid_to_res_gid = {
            gid: sorted(set(tm[slx, sly].flatten()))
            for gid, (slx, sly) in slices.items()
        }

        for sc_gid, res_gids in sc_gid_to_res_gid.items():
            if res_gids[0] < 0:
                sc_gid_to_res_gid[sc_gid] = res_gids[1:]

        if self._gid_map is not None:
            for sc_gid, res_gids in sc_gid_to_res_gid.items():
                sc_gid_to_res_gid[sc_gid] = sorted(
                    self._gid_map[g] for g in res_gids
                )

        logger.info("Pre-loading resource data for Bespoke run... ")
        self._pre_loaded_data = BespokeMultiPlantData(
            self._res_fpath,
            sc_gid_to_hh,
            sc_gid_to_res_gid,
            pre_load_humidity=self._project_points.sam_config_obj.icing,
        )

    def _hh_for_sc_gid(self, sc_gid):
        """Fetch the hh for a given sc_gid"""
        config = self.sam_sys_inputs_with_site_data(sc_gid)
        return int(config["wind_turbine_hub_ht"])

    def _pre_loaded_data_for_sc_gid(self, sc_gid):
        """Pre-load data for a given SC GID, if requested."""
        if self._pre_loaded_data is None:
            return None

        return self._pre_loaded_data.get_preloaded_data_for_gid(sc_gid)

    def _get_bc_for_gid(self, gid):
        """Get the bias correction table trimmed down just for the resource
        pixels corresponding to a single supply curve GID. This can help
        prevent excess memory usage when doing complex bias correction
        distributed to parallel workers.

        Parameters
        ----------
        gid : int
            SC point gid for site to pull bias correction data for

        Returns
        -------
        out : pd.DataFrame | None
            If bias_correct was input, this is just the rows from the larger
            bias correction table that correspond to the SC point gid
        """
        out = self._bias_correct

        if self._bias_correct is not None:
            h5_gids = []
            try:
                scp_kwargs = dict(
                    gid=gid,
                    excl=self._excl_fpath,
                    tm_dset=self._tm_dset,
                    resolution=self._resolution,
                )
                with SupplyCurvePoint(**scp_kwargs) as scp:
                    h5_gids = scp.h5_gid_set
            except EmptySupplyCurvePointError:
                pass

            if self._gid_map is not None:
                h5_gids = [self._gid_map[g] for g in h5_gids]

            mask = self._bias_correct.index.isin(h5_gids)
            out = self._bias_correct[mask]

        return out

    @property
    def outputs(self):
        """Saved outputs for the multi wind plant bespoke optimization. Keys
        are reV supply curve gids and values are BespokeSinglePlant.outputs
        dictionaries.

        Returns
        -------
        dict
        """
        return self._outputs

    @property
    def completed_gids(self):
        """Get a sorted list of completed BespokeSinglePlant gids

        Returns
        -------
        list
        """
        return sorted(list(self.outputs.keys()))

    @property
    def meta(self):
        """Meta data for all completed BespokeSinglePlant objects.

        Returns
        -------
        pd.DataFrame
        """
        meta = [self.outputs[g]["meta"] for g in self.completed_gids]
        if len(self.completed_gids) > 1:
            meta = pd.concat(meta, axis=0)
        else:
            meta = meta[0]
        return meta

    @property
    def slice_lookup(self):
        """Dict | None: Lookup mapping sc_point_gid to exclusion slice."""
        if self._slice_lookup is None and self._inclusion_mask is not None:
            with SupplyCurveExtent(
                self._excl_fpath, resolution=self._resolution
            ) as sc:
                assert self.shape == self._inclusion_mask.shape
                self._slice_lookup = sc.get_slice_lookup(self.gids)

        return self._slice_lookup

    def sam_sys_inputs_with_site_data(self, gid):
        """Update the sam_sys_inputs with site data for the given GID.

        Site data is extracted from the project points DataFrame. Every
        column in the project DataFrame becomes a key in the site_data
        output dictionary.

        Parameters
        ----------
        gid : int
            SC point gid for site to pull site data for.

        Returns
        -------
        dictionary : dict
            SAM system config with extra keys from the project points
            DataFrame.
        """

        gid_idx = self._project_points.index(gid)
        site_data = self._project_points.df.iloc[gid_idx]

        site_sys_inputs = self._project_points[gid][1]
        site_sys_inputs.update(
            {
                k: v
                for k, v in site_data.to_dict().items()
                if not (isinstance(v, float) and np.isnan(v))
            }
        )
        return site_sys_inputs

    def _init_fout(self, out_fpath, sample):
        """Initialize the bespoke output h5 file with meta and time index dsets

        Parameters
        ----------
        out_fpath : str
            Full filepath to an output .h5 file to save Bespoke data to. The
            parent directories will be created if they do not already exist.
        sample : dict
            A single sample BespokeSinglePlant output dict that has been run
            and has output data.
        """
        out_dir = os.path.dirname(out_fpath)
        if not os.path.exists(out_dir):
            create_dirs(out_dir)

        with Outputs(out_fpath, mode="w") as f:
            f._set_meta("meta", self.meta, attrs={})
            ti_dsets = [
                d for d in sample.keys() if d.startswith("time_index-")
            ]
            for dset in ti_dsets:
                f._set_time_index(dset, sample[dset], attrs={})
                f._set_time_index("time_index", sample[dset], attrs={})

    def _collect_out_arr(self, dset, sample):
        """Collect single-plant data arrays into complete arrays with data from
        all BespokeSinglePlant objects.

        Parameters
        ----------
        dset : str
            Dataset to collect, this should be an output dataset present in
            BespokeSinglePlant.outputs
        sample : dict
            A single sample BespokeSinglePlant output dict that has been run
            and has output data.

        Returns
        -------
        full_arr : np.ndarray
            Full data array either 1D for scalar data or 2D for timeseries
            data (n_time, n_plant) for all BespokeSinglePlant objects
        """

        single_arr = sample[dset]

        if isinstance(single_arr, Number):
            shape = (len(self.completed_gids),)
            sample_num = single_arr
        elif isinstance(single_arr, (list, tuple, np.ndarray)):
            shape = (len(single_arr), len(self.completed_gids))
            sample_num = single_arr[0]
        else:
            msg = 'Not writing dataset "{}" of type "{}" to disk.'.format(
                dset, type(single_arr)
            )
            logger.info(msg)
            return None

        if isinstance(sample_num, float):
            dtype = np.float32
        else:
            dtype = type(sample_num)
        full_arr = np.zeros(shape, dtype=dtype)

        # collect data from all wind plants
        logger.info(
            'Collecting dataset "{}" with final shape {}'.format(dset, shape)
        )
        for i, gid in enumerate(self.completed_gids):
            if len(full_arr.shape) == 1:
                full_arr[i] = self.outputs[gid][dset]
            else:
                full_arr[:, i] = self.outputs[gid][dset]

        return full_arr

    def save_outputs(self, out_fpath):
        """Save Bespoke Wind Plant optimization outputs to disk.

        Parameters
        ----------
        out_fpath : str
            Full filepath to an output .h5 file to save Bespoke data to. The
            parent directories will be created if they do not already exist.

        Returns
        -------
        out_fpath : str
            Full filepath to desired .h5 output file, the .h5 extension has
            been added if it was not already present.
        """
        if not out_fpath.endswith(".h5"):
            out_fpath += ".h5"

        if ModuleName.BESPOKE not in out_fpath:
            extension_with_module = "_{}.h5".format(ModuleName.BESPOKE)
            out_fpath = out_fpath.replace(".h5", extension_with_module)

        if not self.completed_gids:
            msg = (
                "No output data found! It is likely that all requested "
                "points are excluded."
            )
            logger.warning(msg)
            warn(msg)
            return out_fpath

        sample = self.outputs[self.completed_gids[0]]
        self._init_fout(out_fpath, sample)

        dsets = [
            d
            for d in sample.keys()
            if not d.startswith("time_index-") and d != "meta"
        ]
        with Outputs(out_fpath, mode="a") as f:
            for dset in dsets:
                full_arr = self._collect_out_arr(dset, sample)
                if full_arr is not None:
                    dset_no_year = dset
                    if parse_year(dset, option="boolean"):
                        year = parse_year(dset)
                        dset_no_year = dset.replace("-{}".format(year), "")

                    attrs = BespokeSinglePlant.OUT_ATTRS.get(dset_no_year, {})
                    attrs = copy.deepcopy(attrs)
                    dtype = attrs.pop("dtype", np.float32)
                    chunks = attrs.pop("chunks", None)
                    try:
                        f.write_dataset(
                            dset, full_arr, dtype, chunks=chunks, attrs=attrs
                        )
                    except Exception as e:
                        msg = 'Failed to write "{}" to disk.'.format(dset)
                        logger.exception(msg)
                        raise OSError(msg) from e

        logger.info("Saved output data to: {}".format(out_fpath))
        return out_fpath

    # pylint: disable=arguments-renamed
    @classmethod
    def run_serial(cls, excl_fpath, res_fpath, tm_dset,
                   sam_sys_inputs, objective_function,
                   capital_cost_function,
                   fixed_operating_cost_function,
                   variable_operating_cost_function,
                   balance_of_system_cost_function,
                   min_spacing='5x', wake_loss_multiplier=1, ga_kwargs=None,
                   output_request=('system_capacity', 'cf_mean'),
                   ws_bins=(0.0, 20.0, 5.0), wd_bins=(0.0, 360.0, 45.0),
                   excl_dict=None, inclusion_mask=None,
                   area_filter_kernel='queen', min_area=None,
                   resolution=64, excl_area=0.0081, data_layers=None,
                   gids=None, exclusion_shape=None, slice_lookup=None,
                   eos_mult_baseline_cap_mw=200, prior_meta=None,
                   gid_map=None, bias_correct=None, pre_loaded_data=None):
        """
        Standalone serial method to run bespoke optimization.
        See BespokeWindPlants docstring for parameter description.

        This method can only take a single sam_sys_inputs... For a spatially
        variant gid-to-config mapping, see the BespokeWindPlants class methods.

        Returns
        -------
        out : dict
            Bespoke outputs keyed by sc point gid
        """

        out = {}
        with SupplyCurveExtent(excl_fpath, resolution=resolution) as sc:
            if gids is None:
                gids = sc.valid_sc_points(tm_dset)
            elif np.issubdtype(type(gids), np.number):
                gids = [gids]
            if slice_lookup is None:
                slice_lookup = sc.get_slice_lookup(gids)
            if exclusion_shape is None:
                exclusion_shape = sc.exclusions.shape

        cls._check_inclusion_mask(inclusion_mask, gids, exclusion_shape)

        # pre-extract handlers so they are not repeatedly initialized
        file_kwargs = {
            "excl_dict": excl_dict,
            "area_filter_kernel": area_filter_kernel,
            "min_area": min_area,
            "h5_handler": MultiYearWindResource,
        }

        with AggFileHandler(excl_fpath, res_fpath, **file_kwargs) as fh:
            n_finished = 0
            for gid in gids:
                gid_inclusions = cls._get_gid_inclusion_mask(
                    inclusion_mask, gid, slice_lookup, resolution=resolution
                )
                try:
                    bsp_plant_out = BespokeSinglePlant.run(
                        gid,
                        fh.exclusions,
                        fh.h5,
                        tm_dset,
                        sam_sys_inputs,
                        objective_function,
                        capital_cost_function,
                        fixed_operating_cost_function,
                        variable_operating_cost_function,
                        balance_of_system_cost_function,
                        min_spacing=min_spacing,
                        wake_loss_multiplier=wake_loss_multiplier,
                        ga_kwargs=ga_kwargs,
                        output_request=output_request,
                        ws_bins=ws_bins,
                        wd_bins=wd_bins,
                        excl_dict=excl_dict,
                        inclusion_mask=gid_inclusions,
                        resolution=resolution,
                        excl_area=excl_area,
                        data_layers=data_layers,
                        exclusion_shape=exclusion_shape,
                        eos_mult_baseline_cap_mw=eos_mult_baseline_cap_mw,
                        prior_meta=prior_meta,
                        gid_map=gid_map,
                        bias_correct=bias_correct,
                        pre_loaded_data=pre_loaded_data,
                        close=False,
                    )

                except EmptySupplyCurvePointError:
                    logger.debug(
                        "SC gid {} is fully excluded or does not "
                        "have any valid source data!".format(gid)
                    )
                except Exception as e:
                    msg = "SC gid {} failed!".format(gid)
                    logger.exception(msg)
                    raise RuntimeError(msg) from e
                else:
                    n_finished += 1
                    logger.debug(
                        "Serial bespoke: "
                        "{} out of {} points complete".format(
                            n_finished, len(gids)
                        )
                    )
                    log_mem(logger)
                    out[gid] = bsp_plant_out

        return out

    def run_parallel(self, max_workers=None):
        """Run the bespoke optimization for many supply curve points in
        parallel.

        Parameters
        ----------
        max_workers : int | None, optional
            Number of cores to run summary on. None is all
            available cpus, by default None

        Returns
        -------
        out : dict
            Bespoke outputs keyed by sc point gid
        """

        logger.info(
            "Running bespoke optimization for points {} through {} "
            "at a resolution of {} on {} cores.".format(
                self.gids[0], self.gids[-1], self._resolution, max_workers
            )
        )

        futures = []
        out = {}
        n_finished = 0
        loggers = [__name__, "reV.supply_curve.point_summary", "reV"]
        with SpawnProcessPool(max_workers=max_workers, loggers=loggers) as exe:
            # iterate through split executions, submitting each to worker
            for gid in self.gids:
                # submit executions and append to futures list
                gid_incl_mask = None
                if self._inclusion_mask is not None:
                    rs, cs = self.slice_lookup[gid]
                    gid_incl_mask = self._inclusion_mask[rs, cs]

                futures.append(exe.submit(
                    self.run_serial,
                    self._excl_fpath,
                    self._res_fpath,
                    self._tm_dset,
                    self.sam_sys_inputs_with_site_data(gid),
                    self._obj_fun,
                    self._cap_cost_fun,
                    self._foc_fun,
                    self._voc_fun,
                    self._bos_fun,
                    self._min_spacing,
                    wake_loss_multiplier=self._wake_loss_multiplier,
                    ga_kwargs=self._ga_kwargs,
                    output_request=self._output_request,
                    ws_bins=self._ws_bins,
                    wd_bins=self._wd_bins,
                    excl_dict=self._excl_dict,
                    inclusion_mask=gid_incl_mask,
                    area_filter_kernel=self._area_filter_kernel,
                    min_area=self._min_area,
                    resolution=self._resolution,
                    excl_area=self._excl_area,
                    data_layers=self._data_layers,
                    gids=gid,
                    exclusion_shape=self.shape,
                    slice_lookup=copy.deepcopy(self.slice_lookup),
                    eos_mult_baseline_cap_mw=self._eos_mult_baseline_cap_mw,
                    prior_meta=self._get_prior_meta(gid),
                    gid_map=self._gid_map,
                    bias_correct=self._get_bc_for_gid(gid),
                    pre_loaded_data=self._pre_loaded_data_for_sc_gid(gid)))

            # gather results
            for future in as_completed(futures):
                n_finished += 1
                out.update(future.result())
                if n_finished % 10 == 0:
                    mem = psutil.virtual_memory()
                    logger.info(
                        "Parallel bespoke futures collected: "
                        "{} out of {}. Memory usage is {:.3f} GB out "
                        "of {:.3f} GB ({:.2f}% utilized).".format(
                            n_finished,
                            len(futures),
                            mem.used / 1e9,
                            mem.total / 1e9,
                            100 * mem.used / mem.total,
                        )
                    )

        return out

    def run(self, out_fpath=None, max_workers=None):
        """Run the bespoke wind plant optimization in serial or parallel.

        Parameters
        ----------
        out_fpath : str, optional
            Path to output file. If ``None``, no output file will
            be written. If the filepath is specified but the module name
            (bespoke) is not included, the module name will get added to
            the output file name. By default, ``None``.
        max_workers : int, optional
            Number of local workers to run on. If ``None``, uses all
            available cores (typically 36). By default, ``None``.

        Returns
        -------
        str | None
            Path to output HDF5 file, or ``None`` if results were not
            written to disk.
        """

        # parallel job distribution test.
        if self._obj_fun == "test":
            return True

        if max_workers == 1:
            slice_lookup = copy.deepcopy(self.slice_lookup)
            for gid in self.gids:
                gid_incl_mask = None
                if self._inclusion_mask is not None:
                    rs, cs = slice_lookup[gid]
                    gid_incl_mask = self._inclusion_mask[rs, cs]

                sam_inputs = self.sam_sys_inputs_with_site_data(gid)
                prior_meta = self._get_prior_meta(gid)
                pre_loaded_data = self._pre_loaded_data_for_sc_gid(gid)
                afk = self._area_filter_kernel
                wlm = self._wake_loss_multiplier
                i_bc = self._get_bc_for_gid(gid)
                ebc = self._eos_mult_baseline_cap_mw

                si = self.run_serial(self._excl_fpath,
                                     self._res_fpath,
                                     self._tm_dset,
                                     sam_inputs,
                                     self._obj_fun,
                                     self._cap_cost_fun,
                                     self._foc_fun,
                                     self._voc_fun,
                                     self._bos_fun,
                                     min_spacing=self._min_spacing,
                                     wake_loss_multiplier=wlm,
                                     ga_kwargs=self._ga_kwargs,
                                     output_request=self._output_request,
                                     ws_bins=self._ws_bins,
                                     wd_bins=self._wd_bins,
                                     excl_dict=self._excl_dict,
                                     inclusion_mask=gid_incl_mask,
                                     area_filter_kernel=afk,
                                     min_area=self._min_area,
                                     resolution=self._resolution,
                                     excl_area=self._excl_area,
                                     data_layers=self._data_layers,
                                     slice_lookup=slice_lookup,
                                     eos_mult_baseline_cap_mw=ebc,
                                     prior_meta=prior_meta,
                                     gid_map=self._gid_map,
                                     bias_correct=i_bc,
                                     gids=gid,
                                     pre_loaded_data=pre_loaded_data)
                self._outputs.update(si)
        else:
            self._outputs = self.run_parallel(max_workers=max_workers)

        if out_fpath is not None:
            out_fpath = self.save_outputs(out_fpath)

        return out_fpath



# -*- coding: utf-8 -*-
# pylint: disable=anomalous-backslash-in-string
"""reV supply curve aggregation framework.

Created on Fri Jun 21 13:24:31 2019

@author: gbuster
"""
import logging
import os
from concurrent.futures import as_completed
from warnings import warn

import numpy as np
import pandas as pd
import psutil
from rex.multi_file_resource import MultiFileResource
from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool

from reV.handlers.exclusions import ExclusionLayers
from reV.supply_curve.aggregation import (
    AbstractAggFileHandler,
    Aggregation,
    BaseAggregation,
)
from reV.supply_curve.exclusions import FrictionMask
from reV.supply_curve.extent import SupplyCurveExtent
from reV.supply_curve.points import GenerationSupplyCurvePoint
from reV.utilities import ResourceMetaField, SupplyCurveField, log_versions
from reV.utilities.exceptions import (
    EmptySupplyCurvePointError,
    FileInputError,
    InputWarning,
    OutputWarning,
)

logger = logging.getLogger(__name__)


class SupplyCurveAggFileHandler(AbstractAggFileHandler):
    """
    Framework to handle aggregation summary context managers:
    - exclusions .h5 file
    - generation .h5 file
    - econ .h5 file (optional)
    - friction surface .h5 file (optional)
    - variable power density .csv (optional)
    """

    def __init__(
        self,
        excl_fpath,
        gen_fpath,
        econ_fpath=None,
        data_layers=None,
        power_density=None,
        excl_dict=None,
        friction_fpath=None,
        friction_dset=None,
        area_filter_kernel="queen",
        min_area=None,
    ):
        """
        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        gen_fpath : str
            Filepath to .h5 reV generation output results.
        econ_fpath : str | None
            Filepath to .h5 reV econ output results. This is optional and only
            used if the lcoe_dset is not present in the gen_fpath file.
        data_layers : None | dict
            Aggregation data layers. Must be a dictionary keyed by data label
            name. Each value must be another dictionary with "dset", "method",
            and "fpath".
        power_density : float | str | None
            Power density in MW/km2 or filepath to variable power
            density file. None will attempt to infer a constant
            power density from the generation meta data technology.
            Variable power density csvs must have "gid" and "power_density"
            columns where gid is the resource gid (typically wtk or nsrdb gid)
            and the power_density column is in MW/km2.
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
        friction_fpath : str | None
            Filepath to friction surface data (cost based exclusions).
            Must be paired with friction_dset. The friction data must be the
            same shape as the exclusions. Friction input creates a new output
            "mean_lcoe_friction" which is the nominal LCOE multiplied by the
            friction data.
        friction_dset : str | None
            Dataset name in friction_fpath for the friction surface data.
            Must be paired with friction_fpath. Must be same shape as
            exclusions.
        area_filter_kernel : str
            Contiguous area filter method to use on final exclusions mask
        min_area : float | None
            Minimum required contiguous area filter in sq-km
        """
        super().__init__(
            excl_fpath,
            excl_dict=excl_dict,
            area_filter_kernel=area_filter_kernel,
            min_area=min_area,
        )

        self._gen = self._open_gen_econ_resource(gen_fpath, econ_fpath)
        # pre-initialize the resource meta data
        _ = self._gen.meta

        self._data_layers = data_layers
        self._power_density = power_density
        self._parse_power_density()

        self._friction_layer = None
        if friction_fpath is not None and friction_dset is not None:
            self._friction_layer = FrictionMask(friction_fpath, friction_dset)

            if not np.all(self._friction_layer.shape == self._excl.shape):
                e = ("Friction layer shape {} must match exclusions shape {}!"
                     .format(self._friction_layer.shape, self._excl.shape))
                logger.error(e)
                raise FileInputError(e)

    @staticmethod
    def _open_gen_econ_resource(gen_fpath, econ_fpath):
        """Open a rex resource file handler for the reV generation and
        (optionally) the reV econ output(s).

        Parameters
        ----------
        gen_fpath : str
            Filepath to .h5 reV generation output results.
        econ_fpath : str | None
            Filepath to .h5 reV econ output results. This is optional and only
            used if the lcoe_dset is not present in the gen_fpath file.

        Returns
        -------
        handler : Resource | MultiFileResource
            Open resource handler initialized with gen_fpath and (optionally)
            econ_fpath.
        """

        handler = None
        is_gen_h5 = isinstance(gen_fpath, str) and gen_fpath.endswith(".h5")
        is_econ_h5 = isinstance(econ_fpath, str) and econ_fpath.endswith(".h5")

        if is_gen_h5 and not is_econ_h5:
            handler = Resource(gen_fpath)
        elif is_gen_h5 and is_econ_h5:
            handler = MultiFileResource(
                [gen_fpath, econ_fpath], check_files=True
            )

        return handler

    def _parse_power_density(self):
        """Parse the power density input. If file, open file handler."""

        if isinstance(self._power_density, str):
            self._pdf = self._power_density

            if self._pdf.endswith(".csv"):
                self._power_density = pd.read_csv(self._pdf)
                if (ResourceMetaField.GID in self._power_density
                        and 'power_density' in self._power_density):
                    self._power_density = \
                        self._power_density.set_index(ResourceMetaField.GID)
                else:
                    msg = ('Variable power density file must include "{}" '
                           'and "power_density" columns, but received: {}'
                           .format(ResourceMetaField.GID,
                                   self._power_density.columns.values))
                    logger.error(msg)
                    raise FileInputError(msg)
            else:
                msg = (
                    "Variable power density file must be csv but received: "
                    "{}".format(self._pdf)
                )
                logger.error(msg)
                raise FileInputError(msg)

    def close(self):
        """Close all file handlers."""
        self._excl.close()
        self._gen.close()
        if self._friction_layer is not None:
            self._friction_layer.close()

    @property
    def gen(self):
        """Get the gen file handler object.

        Returns
        -------
        _gen : Outputs
            reV gen outputs handler object.
        """
        return self._gen

    @property
    def data_layers(self):
        """Get the data layers object.

        Returns
        -------
        _data_layers : dict
            Data layers namespace.
        """
        return self._data_layers

    @property
    def power_density(self):
        """Get the power density object.

        Returns
        -------
        _power_density : float | None | pd.DataFrame
            Constant power density float, None, or opened dataframe with
            (resource) "gid" and "power_density columns".
        """
        return self._power_density

    @property
    def friction_layer(self):
        """Get the friction layer (cost based exclusions).

        Returns
        -------
        friction_layer : None | FrictionMask
            Friction layer with scalar friction values if valid friction inputs
            were entered. Otherwise, None to not apply friction layer.
        """
        return self._friction_layer


class SupplyCurveAggregation(BaseAggregation):
    """SupplyCurveAggregation"""

    def __init__(self, excl_fpath, tm_dset, econ_fpath=None,
                 excl_dict=None, area_filter_kernel='queen', min_area=None,
                 resolution=64, excl_area=None, res_fpath=None, gids=None,
                 pre_extract_inclusions=False, res_class_dset=None,
                 res_class_bins=None, cf_dset='cf_mean-means',
                 lcoe_dset='lcoe_fcr-means', h5_dsets=None, data_layers=None,
                 power_density=None, friction_fpath=None, friction_dset=None,
                 cap_cost_scale=None, recalc_lcoe=True):
        r"""ReV supply curve points aggregation framework.

        ``reV`` supply curve aggregation combines a high-resolution
        (e.g. 90m) exclusion dataset with a (typically) lower resolution
        (e.g. 2km) generation dataset by mapping all data onto the high-
        resolution grid and aggregating it by a large factor (e.g. 64 or
        128). The result is coarsely-gridded data that summarizes
        capacity and generation potential as well as associated
        economics under a particular land access scenario. This module
        can also summarize extra data layers during the aggregation
        process, allowing for complementary land characterization
        analysis.

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions data HDF5 file. The exclusions HDF5
            file should contain the layers specified in `excl_dict`
            and `data_layers`. These layers may also be spread out
            across multiple HDF5 files, in which case this input should
            be a list or tuple of filepaths pointing to the files
            containing the layers. Note that each data layer must be
            uniquely defined (i.e.only appear once and in a single
            input file).
        tm_dset : str
            Dataset name in the `excl_fpath` file containing the
            techmap (exclusions-to-resource mapping data). This data
            layer links the supply curve GID's to the generation GID's
            that are used to evaluate performance metrics such as
            ``mean_cf``.

            .. Important:: This dataset uniquely couples the (typically
              high-resolution) exclusion layers to the (typically
              lower-resolution) resource data. Therefore, a separate
              techmap must be used for every unique combination of
              resource and exclusion coordinates.

            .. Note:: If executing ``reV`` from the command line, you
              can specify a name that is not in the exclusions HDF5
              file, and ``reV`` will calculate the techmap for you. Note
              however that computing the techmap and writing it to the
              exclusion HDF5 file is a blocking operation, so you may
              only run a single ``reV`` aggregation step at a time this
              way.

        econ_fpath : str, optional
            Filepath to HDF5 file with ``reV`` econ output results
            containing an `lcoe_dset` dataset. If ``None``, `lcoe_dset`
            should be a dataset in the `gen_fpath` HDF5 file that
            aggregation is executed on.

            .. Note:: If executing ``reV`` from the command line, this
              input can be set to ``"PIPELINE"`` to parse the output
              from one of these preceding pipeline steps:
              ``multi-year``, ``collect``, or ``generation``. However,
              note that duplicate executions of any of these commands
              within the pipeline may invalidate this parsing, meaning
              the `econ_fpath` input will have to be specified manually.

            By default, ``None``.
        excl_dict : dict | None
            Dictionary of exclusion keyword arguments of the format
            ``{layer_dset_name: {kwarg: value}}``, where
            ``layer_dset_name`` is a dataset in the exclusion h5 file
            and the ``kwarg: value`` pair is a keyword argument to
            the :class:`reV.supply_curve.exclusions.LayerMask` class.
            For example::

                excl_dict = {
                    "typical_exclusion": {
                        "exclude_values": 255,
                    },
                    "another_exclusion": {
                        "exclude_values": [2, 3],
                        "weight": 0.5
                    },
                    "exclusion_with_nodata": {
                        "exclude_range": [10, 100],
                        "exclude_nodata": True,
                        "nodata_value": -1
                    },
                    "partial_setback": {
                        "use_as_weights": True
                    },
                    "height_limit": {
                        "exclude_range": [0, 200]
                    },
                    "slope": {
                        "include_range": [0, 20]
                    },
                    "developable_land": {
                        "force_include_values": 42
                    },
                    "more_developable_land": {
                        "force_include_range": [5, 10]
                    },
                    "viewsheds": {
                        "exclude_values": 1,
                        "extent": {
                            "layer": "federal_parks",
                            "include_range": [1, 5]
                        }
                    }
                    ...
                }

            Note that all the keys given in this dictionary should be
            datasets of the `excl_fpath` file. If ``None`` or empty
            dictionary, no exclusions are applied. By default, ``None``.
        area_filter_kernel : {"queen", "rook"}, optional
            Contiguous area filter method to use on final exclusions
            mask. The filters are defined as::

                # Queen:     # Rook:
                [[1,1,1],    [[0,1,0],
                 [1,1,1],     [1,1,1],
                 [1,1,1]]     [0,1,0]]

            These filters define how neighboring pixels are "connected".
            Once pixels in the final exclusion layer are connected, the
            area of each resulting cluster is computed and compared
            against the `min_area` input. Any cluster with an area
            less than `min_area` is excluded from the final mask.
            This argument has no effect if `min_area` is ``None``.
            By default, ``"queen"``.
        min_area : float, optional
            Minimum area (in km\ :sup:`2`) required to keep an isolated
            cluster of (included) land within the resulting exclusions
            mask. Any clusters of land with areas less than this value
            will be marked as exclusions. See the documentation for
            `area_filter_kernel` for an explanation of how the area of
            each land cluster is computed. If ``None``, no area
            filtering is performed. By default, ``None``.
        resolution : int, optional
            Supply Curve resolution. This value defines how many pixels
            are in a single side of a supply curve cell. For example,
            a value of ``64`` would generate a supply curve where the
            side of each supply curve cell is ``64x64`` exclusion
            pixels. By default, ``64``.
        excl_area : float, optional
            Area of a single exclusion mask pixel (in km\ :sup:`2`).
            If ``None``, this value will be inferred from the profile
            transform attribute in `excl_fpath`. By default, ``None``.
        res_fpath : str, optional
            Filepath to HDF5 resource file (e.g. WTK or NSRDB). This
            input is required if techmap dset is to be created or if the
            ``gen_fpath`` input to the ``summarize`` or ``run`` methods
            is ``None``. By default, ``None``.
        gids : list, optional
            List of supply curve point gids to get summary for. If you
            would like to obtain all available ``reV`` supply curve
            points to run, you can use the
            :class:`reV.supply_curve.extent.SupplyCurveExtent` class
            like so::

                import pandas as pd
                from reV.supply_curve.extent import SupplyCurveExtent

                excl_fpath = "..."
                resolution = ...
                tm_dset = "..."
                with SupplyCurveExtent(excl_fpath, resolution) as sc:
                    gids = sc.valid_sc_points(tm_dset).tolist()
                ...

            If ``None``, supply curve aggregation is computed for all
            gids in the supply curve extent. By default, ``None``.
        pre_extract_inclusions : bool, optional
            Optional flag to pre-extract/compute the inclusion mask from
            the `excl_dict` input. It is typically faster to compute
            the inclusion mask on the fly with parallel workers.
            By default, ``False``.
        res_class_dset : str, optional
            Name of dataset in the ``reV`` generation HDF5 output file
            containing resource data. If ``None``, no aggregated
            resource classification is performed (i.e. no ``mean_res``
            output), and the `res_class_bins` is ignored.
            By default, ``None``.
        res_class_bins : list, optional
            Optional input to perform separate aggregations for various
            resource data ranges. If ``None``, only a single aggregation
            per supply curve point is performed. Otherwise, this input
            should be a list of floats or ints representing the resource
            bin boundaries. One aggregation per resource value range is
            computed, and only pixels within the given resource range
            are aggregated. By default, ``None``.
        cf_dset : str, optional
            Dataset name from the ``reV`` generation HDF5 output file
            containing a 1D dataset of mean capacity factor values. This
            dataset will be mapped onto the high-resolution grid and
            used to compute the mean capacity factor for non-excluded
            area. By default, ``"cf_mean-means"``.
        lcoe_dset : str, optional
            Dataset name from the ``reV`` generation HDF5 output file
            containing a 1D dataset of mean LCOE values. This
            dataset will be mapped onto the high-resolution grid and
            used to compute the mean LCOE for non-excluded area, but
            only if the LCOE is not re-computed during processing (see
            the `recalc_lcoe` input for more info).
            By default, ``"lcoe_fcr-means"``.
        h5_dsets : list, optional
            Optional list of additional datasets from the ``reV``
            generation/econ HDF5 output file to aggregate. If ``None``,
            no extra datasets are aggregated.

            .. WARNING:: This input is meant for passing through 1D
               datasets. If you specify a 2D or higher-dimensional
               dataset, you may run into memory errors. If you wish to
               aggregate 2D datasets, see the rep-profiles module.

            By default, ``None``.
        data_layers : dict, optional
            Dictionary of aggregation data layers of the format::

                data_layers = {
                    "output_layer_name": {
                        "dset": "layer_name",
                        "method": "mean",
                        "fpath": "/path/to/data.h5"
                    },
                    "another_output_layer_name": {
                        "dset": "input_layer_name",
                        "method": "mode",
                        # optional "fpath" key omitted
                    },
                    ...
                }

            The ``"output_layer_name"`` is the column name under which
            the aggregated data will appear in the output CSV file. The
            ``"output_layer_name"`` does not have to match the ``dset``
            input value. The latter should match the layer name in the
            HDF5 from which the data to aggregate should be pulled. The
            ``method`` should be one of
            ``{"mode", "mean", "min", "max", "sum", "category"}``,
            describing how the high-resolution data should be aggregated
            for each supply curve point. ``fpath`` is an optional key
            that can point to an HDF5 file containing the layer data. If
            left out, the data is assumed to exist in the file(s)
            specified by the `excl_fpath` input. If ``None``, no data
            layer aggregation is performed. By default, ``None``
        power_density : float | str, optional
            Power density value (in MW/km\ :sup:`2`) or filepath to
            variable power density CSV file containing the following
            columns:

                - ``gid`` : resource gid (typically wtk or nsrdb gid)
                - ``power_density`` : power density value (in
                  MW/km\ :sup:`2`)

            If ``None``, a constant power density is inferred from the
            generation meta data technology. By default, ``None``.
        friction_fpath : str, optional
            Filepath to friction surface data (cost based exclusions).
            Must be paired with the `friction_dset` input below. The
            friction data must be the same shape as the exclusions.
            Friction input creates a new output column
            ``"mean_lcoe_friction"`` which is the nominal LCOE
            multiplied by the friction data. If ``None``, no friction
            data is aggregated. By default, ``None``.
        friction_dset : str, optional
            Dataset name in friction_fpath for the friction surface
            data. Must be paired with the `friction_fpath` above. If
            ``None``, no friction data is aggregated.
            By default, ``None``.
        cap_cost_scale : str, optional
            Optional LCOE scaling equation to implement "economies of
            scale". Equations must be in python string format and must
            return a scalar value to multiply the capital cost by.
            Independent variables in the equation should match the names
            of the columns in the ``reV`` supply curve aggregation
            output table (see the documentation of
            :class:`~reV.supply_curve.sc_aggregation.SupplyCurveAggregation`
            for details on available outputs). If ``None``, no economies
            of scale are applied. By default, ``None``.
        recalc_lcoe : bool, optional
            Flag to re-calculate the LCOE from the multi-year mean
            capacity factor and annual energy production data. This
            requires several datasets to be aggregated in the h5_dsets
            input:

                - ``system_capacity``
                - ``fixed_charge_rate``
                - ``capital_cost``
                - ``fixed_operating_cost``
                - ``variable_operating_cost``

            If any of these datasets are missing from the ``reV``
            generation HDF5 output, or if `recalc_lcoe` is set to
            ``False``, the mean LCOE will be computed from the data
            stored under the `lcoe_dset` instead. By default, ``True``.

        Examples
        --------
        Standard outputs:

        sc_gid : int
            Unique supply curve gid. This is the enumerated supply curve
            points, which can have overlapping geographic locations due
            to different resource bins at the same geographic SC point.
        res_gids : list
            Stringified list of resource gids (e.g. original WTK or
            NSRDB resource GIDs) corresponding to each SC point.
        gen_gids : list
            Stringified list of generation gids (e.g. GID in the reV
            generation output, which corresponds to the reV project
            points and not necessarily the resource GIDs).
        gid_counts : list
            Stringified list of the sum of inclusion scalar values
            corresponding to each `gen_gid` and `res_gid`, where 1 is
            included, 0 is excluded, and 0.7 is included with 70 percent
            of available land. Each entry in this list is associated
            with the corresponding entry in the `gen_gids` and
            `res_gids` lists.
        n_gids : int
            Total number of included pixels. This is a boolean sum and
            considers partial inclusions to be included (e.g. 1).
        mean_cf : float
            Mean capacity factor of each supply curve point (the
            arithmetic mean is weighted by the inclusion layer)
            (unitless).
        mean_lcoe : float
            Mean LCOE of each supply curve point (the arithmetic mean is
            weighted by the inclusion layer). Units match the reV econ
            output ($/MWh). By default, the LCOE is re-calculated using
            the multi-year mean capacity factor and annual energy
            production. This requires several datasets to be aggregated
            in the h5_dsets input: ``fixed_charge_rate``,
            ``capital_cost``,
            ``fixed_operating_cost``, ``annual_energy_production``, and
            ``variable_operating_cost``. This recalc behavior can be
            disabled by setting ``recalc_lcoe=False``.
        mean_res : float
            Mean resource, the resource dataset to average is provided
            by the user in `res_class_dset`. The arithmetic mean is
            weighted by the inclusion layer.
        capacity : float
            Total capacity of each supply curve point (MW). Units are
            contingent on the `power_density` input units of MW/km2.
        area_sq_km : float
            Total included area for each supply curve point in km2. This
            is based on the nominal area of each exclusion pixel which
            by default is calculated from the exclusion profile
            attributes. The NREL reV default is 0.0081 km2 pixels
            (90m x 90m). The area sum considers partial inclusions.
        latitude : float
            Supply curve point centroid latitude coordinate, in degrees
            (does not consider exclusions).
        longitude : float
            Supply curve point centroid longitude coordinate, in degrees
            (does not consider exclusions).
        country : str
            Country of the supply curve point based on the most common
            country of the associated resource meta data. Does not
            consider exclusions.
        state : str
            State of the supply curve point based on the most common
            state of the associated resource meta data. Does not
            consider exclusions.
        county : str
            County of the supply curve point based on the most common
            county of the associated resource meta data. Does not
            consider exclusions.
        elevation : float
            Mean elevation of the supply curve point based on the mean
            elevation of the associated resource meta data. Does not
            consider exclusions.
        timezone : int
            UTC offset of local timezone based on the most common
            timezone of the associated resource meta data. Does not
            consider exclusions.
        sc_point_gid : int
            Spatially deterministic supply curve point gid. Duplicate
            `sc_point_gid` values can exist due to resource binning.
        sc_row_ind : int
            Row index of the supply curve point in the aggregated
            exclusion grid.
        sc_col_ind : int
            Column index of the supply curve point in the aggregated
            exclusion grid
        res_class : int
            Resource class for the supply curve gid. Each geographic
            supply curve point (`sc_point_gid`) can have multiple
            resource classes associated with it, resulting in multiple
            supply curve gids (`sc_gid`) associated with the same
            spatially deterministic supply curve point.


        Optional outputs:

        mean_friction : float
            Mean of the friction data provided in 'friction_fpath' and
            'friction_dset'. The arithmetic mean is weighted by boolean
            inclusions and considers partial inclusions to be included.
        mean_lcoe_friction : float
            Mean of the nominal LCOE multiplied by mean_friction value.
        mean_{dset} : float
            Mean input h5 dataset(s) provided by the user in 'h5_dsets'.
            These mean calculations are weighted by the partial
            inclusion layer.
        data_layers : float | int | str | dict
            Requested data layer aggregations, each data layer must be
            the same shape as the exclusion layers.

                - mode: int | str
                    Most common value of a given data layer after
                    applying the boolean inclusion mask.
                - mean : float
                    Arithmetic mean value of a given data layer weighted
                    by the scalar inclusion mask (considers partial
                    inclusions).
                - min : float | int
                    Minimum value of a given data layer after applying
                    the boolean inclusion mask.
                - max : float | int
                    Maximum value of a given data layer after applying
                    the boolean inclusion mask.
                - sum : float
                    Sum of a given data layer weighted by the scalar
                    inclusion mask (considers partial inclusions).
                - category : dict
                    Dictionary mapping the unique values in the
                    `data_layer` to the sum of inclusion scalar values
                    associated with all pixels with that unique value.
        """
        log_versions(logger)
        logger.info("Initializing SupplyCurveAggregation...")
        logger.debug("Exclusion filepath: {}".format(excl_fpath))
        logger.debug("Exclusion dict: {}".format(excl_dict))

        super().__init__(
            excl_fpath,
            tm_dset,
            excl_dict=excl_dict,
            area_filter_kernel=area_filter_kernel,
            min_area=min_area,
            resolution=resolution,
            excl_area=excl_area,
            res_fpath=res_fpath,
            gids=gids,
            pre_extract_inclusions=pre_extract_inclusions,
        )

        self._econ_fpath = econ_fpath
        self._res_class_dset = res_class_dset
        self._res_class_bins = self._convert_bins(res_class_bins)
        self._cf_dset = cf_dset
        self._lcoe_dset = lcoe_dset
        self._h5_dsets = h5_dsets
        self._cap_cost_scale = cap_cost_scale
        self._power_density = power_density
        self._friction_fpath = friction_fpath
        self._friction_dset = friction_dset
        self._data_layers = data_layers
        self._recalc_lcoe = recalc_lcoe

        logger.debug("Resource class bins: {}".format(self._res_class_bins))

        if self._power_density is None:
            msg = (
                "Supply curve aggregation power density not specified. "
                "Will try to infer based on lookup table: {}".format(
                    GenerationSupplyCurvePoint.POWER_DENSITY
                )
            )
            logger.warning(msg)
            warn(msg, InputWarning)

        self._check_data_layers()

    def _check_data_layers(
        self, methods=("mean", "max", "min", "mode", "sum", "category")
    ):
        """Run pre-flight checks on requested aggregation data layers.

        Parameters
        ----------
        methods : list | tuple
            Data layer aggregation methods that are available to the user.
        """

        if self._data_layers is not None:
            logger.debug("Checking data layers...")

            with ExclusionLayers(self._excl_fpath) as f:
                shape_base = f.shape

            for k, v in self._data_layers.items():
                if "dset" not in v:
                    raise KeyError(
                        'Data aggregation "dset" data layer "{}" '
                        "must be specified.".format(k)
                    )
                if "method" not in v:
                    raise KeyError(
                        'Data aggregation "method" data layer "{}" '
                        "must be specified.".format(k)
                    )
                if v["method"].lower() not in methods:
                    raise ValueError(
                        "Cannot recognize data layer agg method: "
                        '"{}". Can only do: {}.'.format(v["method"], methods)
                    )
                if "fpath" in v:
                    with ExclusionLayers(v["fpath"]) as f:
                        try:
                            mismatched_shapes = any(f.shape != shape_base)
                        except TypeError:
                            mismatched_shapes = f.shape != shape_base
                        if mismatched_shapes:
                            msg = (
                                'Data shape of data layer "{}" is {}, '
                                "which does not match the baseline "
                                "exclusions shape {}.".format(
                                    k, f.shape, shape_base
                                )
                            )
                            raise FileInputError(msg)

        logger.debug("Finished checking data layers.")

    @staticmethod
    def _get_res_gen_lcoe_data(
        gen, res_class_dset, res_class_bins, lcoe_dset
    ):
        """Extract the basic resource / generation / lcoe data to be used in
        the aggregation process.

        Parameters
        ----------
        gen : Resource | MultiFileResource
            Open rex resource handler initialized from gen_fpath and
            (optionally) econ_fpath.
        res_class_dset : str | None
            Dataset in the generation file dictating resource classes.
            None if no resource classes.
        res_class_bins : list | None
            List of two-entry lists dictating the resource class bins.
            None if no resource classes.
        lcoe_dset : str
            Dataset name from f_gen containing LCOE mean values.

        Returns
        -------
        res_data : np.ndarray | None
            Extracted resource data from res_class_dset
        res_class_bins : list
            List of resouce class bin ranges.
        lcoe_data : np.ndarray | None
            LCOE data extracted from lcoe_dset in gen
        """

        dset_list = (res_class_dset, lcoe_dset)
        gen_dsets = [] if gen is None else gen.datasets
        labels = ("res_class_dset", "lcoe_dset")
        temp = [None, None]

        if isinstance(gen, Resource):
            source_fps = [gen.h5_file]
        elif isinstance(gen, MultiFileResource):
            source_fps = gen._h5_files
        else:
            msg = 'Did not recognize gen object input of type "{}": {}'.format(
                type(gen), gen
            )
            logger.error(msg)
            raise TypeError(msg)

        for i, dset in enumerate(dset_list):
            if dset in gen_dsets:
                _warn_about_large_datasets(gen, dset)
                temp[i] = gen[dset]
            elif dset not in gen_dsets and dset is not None:
                w = (
                    'Could not find "{}" input as "{}" in source files: {}. '
                    "Available datasets: {}".format(
                        labels[i], dset, source_fps, gen_dsets
                    )
                )
                logger.warning(w)
                warn(w, OutputWarning)

        res_data, lcoe_data = temp

        if res_class_dset is None or res_class_bins is None:
            res_class_bins = [None]

        return res_data, res_class_bins, lcoe_data

    @staticmethod
    def _get_extra_dsets(gen, h5_dsets):
        """Extract extra ancillary datasets to be used in the aggregation
        process

        Parameters
        ----------
        gen : Resource | MultiFileResource
            Open rex resource handler initialized from gen_fpath and
            (optionally) econ_fpath.
        h5_dsets : list | None
            Optional list of additional datasets from the source h5 gen/econ
            files to aggregate.

        Returns
        -------
        h5_dsets_data : dict | None
            If additional h5_dsets are requested, this will be a dictionary
            keyed by the h5 dataset names. The corresponding values will be
            the extracted arrays from the h5 files.
        """

        # look for the datasets required by the LCOE re-calculation and make
        # lists of the missing datasets
        gen_dsets = [] if gen is None else gen.datasets
        lcoe_recalc_req = ('fixed_charge_rate',
                           'capital_cost',
                           'fixed_operating_cost',
                           'variable_operating_cost',
                           'system_capacity')
        missing_lcoe_source = [k for k in lcoe_recalc_req
                               if k not in gen_dsets]

        if isinstance(gen, Resource):
            source_fps = [gen.h5_file]
        elif isinstance(gen, MultiFileResource):
            source_fps = gen._h5_files
        else:
            msg = 'Did not recognize gen object input of type "{}": {}'.format(
                type(gen), gen
            )
            logger.error(msg)
            raise TypeError(msg)

        h5_dsets_data = None
        if h5_dsets is not None:

            if not isinstance(h5_dsets, (list, tuple)):
                e = (
                    "Additional h5_dsets argument must be a list or tuple "
                    "but received: {} {}".format(type(h5_dsets), h5_dsets)
                )
                logger.error(e)
                raise TypeError(e)

            missing_h5_dsets = [k for k in h5_dsets if k not in gen_dsets]
            if any(missing_h5_dsets):
                msg = (
                    'Could not find requested h5_dsets "{}" in '
                    "source files: {}. Available datasets: {}".format(
                        missing_h5_dsets, source_fps, gen_dsets
                    )
                )
                logger.error(msg)
                raise FileInputError(msg)

            h5_dsets_data = {dset: gen[dset] for dset in h5_dsets}

        if any(missing_lcoe_source):
            msg = (
                "Could not find the datasets in the gen source file that "
                "are required to re-calculate the multi-year LCOE. If you "
                "are running a multi-year job, it is strongly suggested "
                "you pass through these datasets to re-calculate the LCOE "
                "from the multi-year mean CF: {}".format(missing_lcoe_source)
            )
            logger.warning(msg)
            warn(msg, InputWarning)

        return h5_dsets_data

    @classmethod
    def run_serial(
        cls,
        excl_fpath,
        gen_fpath,
        tm_dset,
        gen_index,
        econ_fpath=None,
        excl_dict=None,
        inclusion_mask=None,
        area_filter_kernel="queen",
        min_area=None,
        resolution=64,
        gids=None,
        args=None,
        res_class_dset=None,
        res_class_bins=None,
        cf_dset="cf_mean-means",
        lcoe_dset="lcoe_fcr-means",
        h5_dsets=None,
        data_layers=None,
        power_density=None,
        friction_fpath=None,
        friction_dset=None,
        excl_area=None,
        cap_cost_scale=None,
        recalc_lcoe=True,
    ):
        """Standalone method to create agg summary - can be parallelized.

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        gen_fpath : str
            Filepath to .h5 reV generation output results.
        tm_dset : str
            Dataset name in the exclusions file containing the
            exclusions-to-resource mapping data.
        gen_index : np.ndarray
            Array of generation gids with array index equal to resource gid.
            Array value is -1 if the resource index was not used in the
            generation run.
        econ_fpath : str | None
            Filepath to .h5 reV econ output results. This is optional and only
            used if the lcoe_dset is not present in the gen_fpath file.
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
        inclusion_mask : np.ndarray | dict | optional
            2D array pre-extracted inclusion mask where 1 is included and 0 is
            excluded. This must be either match the full exclusion shape or
            be a dict lookup of single-sc-point exclusion masks corresponding
            to the gids input and keyed by gids, by default None which will
            calculate exclusions on the fly for each sc point.
        area_filter_kernel : str
            Contiguous area filter method to use on final exclusions mask
        min_area : float | None
            Minimum required contiguous area filter in sq-km
        resolution : int | None
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead.
        gids : list | None
            List of supply curve point gids to get summary for (can use to
            subset if running in parallel), or None for all gids in the SC
            extent, by default None
        args : list | None
            List of positional args for sc_point_method
        res_class_dset : str | None
            Dataset in the generation file dictating resource classes.
            None if no resource classes.
        res_class_bins : list | None
            List of two-entry lists dictating the resource class bins.
            None if no resource classes.
        cf_dset : str
            Dataset name from f_gen containing capacity factor mean values.
        lcoe_dset : str
            Dataset name from f_gen containing LCOE mean values.
        h5_dsets : list | None
            Optional list of additional datasets from the source h5 gen/econ
            files to aggregate.
        data_layers : None | dict
            Aggregation data layers. Must be a dictionary keyed by data label
            name. Each value must be another dictionary with "dset", "method",
            and "fpath".
        power_density : float | str | None
            Power density in MW/km2 or filepath to variable power
            density file. None will attempt to infer a constant
            power density from the generation meta data technology.
            Variable power density csvs must have "gid" and "power_density"
            columns where gid is the resource gid (typically wtk or nsrdb gid)
            and the power_density column is in MW/km2.
        friction_fpath : str | None
            Filepath to friction surface data (cost based exclusions).
            Must be paired with friction_dset. The friction data must be the
            same shape as the exclusions. Friction input creates a new output
            "mean_lcoe_friction" which is the nominal LCOE multiplied by the
            friction data.
        friction_dset : str | None
            Dataset name in friction_fpath for the friction surface data.
            Must be paired with friction_fpath. Must be same shape as
            exclusions.
        excl_area : float | None, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath, by default None
        cap_cost_scale : str | None
            Optional LCOE scaling equation to implement "economies of scale".
            Equations must be in python string format and return a scalar
            value to multiply the capital cost by. Independent variables in
            the equation should match the names of the columns in the reV
            supply curve aggregation table.
        recalc_lcoe : bool
            Flag to re-calculate the LCOE from the multi-year mean capacity
            factor and annual energy production data. This requires several
            datasets to be aggregated in the h5_dsets input: system_capacity,
            fixed_charge_rate, capital_cost, fixed_operating_cost,
            and variable_operating_cost.

        Returns
        -------
        summary : list
            List of dictionaries, each being an SC point summary.
        """
        summary = []

        with SupplyCurveExtent(excl_fpath, resolution=resolution) as sc:
            exclusion_shape = sc.exclusions.shape
            if gids is None:
                gids = sc.valid_sc_points(tm_dset)
            elif np.issubdtype(type(gids), np.number):
                gids = [gids]

            slice_lookup = sc.get_slice_lookup(gids)

        logger.debug(
            "Starting SupplyCurveAggregation serial with "
            "supply curve {} gids".format(len(gids))
        )

        cls._check_inclusion_mask(inclusion_mask, gids, exclusion_shape)

        # pre-extract handlers so they are not repeatedly initialized
        file_kwargs = {
            "econ_fpath": econ_fpath,
            "data_layers": data_layers,
            "power_density": power_density,
            "excl_dict": excl_dict,
            "area_filter_kernel": area_filter_kernel,
            "min_area": min_area,
            "friction_fpath": friction_fpath,
            "friction_dset": friction_dset,
        }
        with SupplyCurveAggFileHandler(
            excl_fpath, gen_fpath, **file_kwargs
        ) as fh:
            temp = cls._get_res_gen_lcoe_data(
                fh.gen, res_class_dset, res_class_bins, lcoe_dset
            )
            res_data, res_class_bins, lcoe_data = temp
            h5_dsets_data = cls._get_extra_dsets(fh.gen, h5_dsets)

            n_finished = 0
            for gid in gids:
                gid_inclusions = cls._get_gid_inclusion_mask(
                    inclusion_mask, gid, slice_lookup, resolution=resolution
                )

                for ri, res_bin in enumerate(res_class_bins):
                    try:
                        pointsum = GenerationSupplyCurvePoint.summarize(
                            gid,
                            fh.exclusions,
                            fh.gen,
                            tm_dset,
                            gen_index,
                            res_class_dset=res_data,
                            res_class_bin=res_bin,
                            cf_dset=cf_dset,
                            lcoe_dset=lcoe_data,
                            h5_dsets=h5_dsets_data,
                            data_layers=fh.data_layers,
                            resolution=resolution,
                            exclusion_shape=exclusion_shape,
                            power_density=fh.power_density,
                            args=args,
                            excl_dict=excl_dict,
                            inclusion_mask=gid_inclusions,
                            excl_area=excl_area,
                            close=False,
                            friction_layer=fh.friction_layer,
                            cap_cost_scale=cap_cost_scale,
                            recalc_lcoe=recalc_lcoe,
                        )

                    except EmptySupplyCurvePointError:
                        logger.debug("SC point {} is empty".format(gid))
                    else:
                        pointsum['res_class'] = ri

                        summary.append(pointsum)
                        logger.debug(
                            "Serial aggregation completed gid {}: "
                            "{} out of {} points complete".format(
                                gid, n_finished, len(gids)
                            )
                        )

                n_finished += 1

        return summary

    def run_parallel(
        self, gen_fpath, args=None, max_workers=None, sites_per_worker=100
    ):
        """Get the supply curve points aggregation summary using futures.

        Parameters
        ----------
        gen_fpath : str
            Filepath to .h5 reV generation output results.
        args : tuple | list | None
            List of summary arguments to include. None defaults to all
            available args defined in the class attr.
        max_workers : int | None, optional
            Number of cores to run summary on. None is all
            available cpus, by default None
        sites_per_worker : int
            Number of sc_points to summarize on each worker, by default 100

        Returns
        -------
        summary : list
            List of dictionaries, each being an SC point summary.
        """

        gen_index = self._parse_gen_index(gen_fpath)
        chunks = int(np.ceil(len(self.gids) / sites_per_worker))
        chunks = np.array_split(self.gids, chunks)

        logger.info(
            "Running supply curve point aggregation for "
            "points {} through {} at a resolution of {} "
            "on {} cores in {} chunks.".format(
                self.gids[0],
                self.gids[-1],
                self._resolution,
                max_workers,
                len(chunks),
            )
        )

        slice_lookup = None
        if self._inclusion_mask is not None:
            with SupplyCurveExtent(
                self._excl_fpath, resolution=self._resolution
            ) as sc:
                assert sc.exclusions.shape == self._inclusion_mask.shape
                slice_lookup = sc.get_slice_lookup(self.gids)

        futures = []
        summary = []
        n_finished = 0
        loggers = [__name__, "reV.supply_curve.point_summary", "reV"]
        with SpawnProcessPool(max_workers=max_workers, loggers=loggers) as exe:
            # iterate through split executions, submitting each to worker
            for gid_set in chunks:
                # submit executions and append to futures list
                chunk_incl_masks = None
                if self._inclusion_mask is not None:
                    chunk_incl_masks = {}
                    for gid in gid_set:
                        rs, cs = slice_lookup[gid]
                        chunk_incl_masks[gid] = self._inclusion_mask[rs, cs]

                futures.append(
                    exe.submit(
                        self.run_serial,
                        self._excl_fpath,
                        gen_fpath,
                        self._tm_dset,
                        gen_index,
                        econ_fpath=self._econ_fpath,
                        excl_dict=self._excl_dict,
                        inclusion_mask=chunk_incl_masks,
                        res_class_dset=self._res_class_dset,
                        res_class_bins=self._res_class_bins,
                        cf_dset=self._cf_dset,
                        lcoe_dset=self._lcoe_dset,
                        h5_dsets=self._h5_dsets,
                        data_layers=self._data_layers,
                        resolution=self._resolution,
                        power_density=self._power_density,
                        friction_fpath=self._friction_fpath,
                        friction_dset=self._friction_dset,
                        area_filter_kernel=self._area_filter_kernel,
                        min_area=self._min_area,
                        gids=gid_set,
                        args=args,
                        excl_area=self._excl_area,
                        cap_cost_scale=self._cap_cost_scale,
                        recalc_lcoe=self._recalc_lcoe,
                    )
                )

            # gather results
            for future in as_completed(futures):
                n_finished += 1
                summary += future.result()
                if n_finished % 10 == 0:
                    mem = psutil.virtual_memory()
                    logger.info(
                        "Parallel aggregation futures collected: "
                        "{} out of {}. Memory usage is {:.3f} GB out "
                        "of {:.3f} GB ({:.2f}% utilized).".format(
                            n_finished,
                            len(chunks),
                            mem.used / 1e9,
                            mem.total / 1e9,
                            100 * mem.used / mem.total,
                        )
                    )

        return summary

    @staticmethod
    def _convert_bins(bins):
        """Convert a list of floats or ints to a list of two-entry bin bounds.

        Parameters
        ----------
        bins : list | None
            List of floats or ints (bin edges) to convert to list of two-entry
            bin boundaries or list of two-entry bind boundaries in final format

        Returns
        -------
        bins : list
            List of two-entry bin boundaries
        """

        if bins is None:
            return None

        type_check = [isinstance(x, (list, tuple)) for x in bins]

        if all(type_check):
            return bins

        if any(type_check):
            raise TypeError(
                "Resource class bins has inconsistent "
                "entry type: {}".format(bins)
            )

        bbins = []
        for i, b in enumerate(sorted(bins)):
            if i < len(bins) - 1:
                bbins.append([b, bins[i + 1]])

        return bbins

    @staticmethod
    def _summary_to_df(summary):
        """Convert the agg summary list to a DataFrame.

        Parameters
        ----------
        summary : list
            List of dictionaries, each being an SC point summary.

        Returns
        -------
        summary : DataFrame
            Summary of the SC points.
        """
        summary = pd.DataFrame(summary)
        sort_by = [x for x in (SupplyCurveField.SC_POINT_GID, 'res_class')
                   if x in summary]
        summary = summary.sort_values(sort_by)
        summary = summary.reset_index(drop=True)
        summary.index.name = SupplyCurveField.SC_GID

        return summary

    def summarize(
        self, gen_fpath, args=None, max_workers=None, sites_per_worker=100
    ):
        """
        Get the supply curve points aggregation summary

        Parameters
        ----------
        gen_fpath : str
            Filepath to .h5 reV generation output results.
        args : tuple | list | None
            List of summary arguments to include. None defaults to all
            available args defined in the class attr.
        max_workers : int | None, optional
            Number of cores to run summary on. None is all
            available cpus, by default None
        sites_per_worker : int
            Number of sc_points to summarize on each worker, by default 100

        Returns
        -------
        summary : list
            List of dictionaries, each being an SC point summary.
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        if max_workers == 1:
            gen_index = self._parse_gen_index(gen_fpath)
            afk = self._area_filter_kernel
            summary = self.run_serial(
                self._excl_fpath,
                gen_fpath,
                self._tm_dset,
                gen_index,
                econ_fpath=self._econ_fpath,
                excl_dict=self._excl_dict,
                inclusion_mask=self._inclusion_mask,
                res_class_dset=self._res_class_dset,
                res_class_bins=self._res_class_bins,
                cf_dset=self._cf_dset,
                lcoe_dset=self._lcoe_dset,
                h5_dsets=self._h5_dsets,
                data_layers=self._data_layers,
                resolution=self._resolution,
                power_density=self._power_density,
                friction_fpath=self._friction_fpath,
                friction_dset=self._friction_dset,
                area_filter_kernel=afk,
                min_area=self._min_area,
                gids=self.gids,
                args=args,
                excl_area=self._excl_area,
                cap_cost_scale=self._cap_cost_scale,
                recalc_lcoe=self._recalc_lcoe,
            )
        else:
            summary = self.run_parallel(
                gen_fpath=gen_fpath,
                args=args,
                max_workers=max_workers,
                sites_per_worker=sites_per_worker,
            )

        if not any(summary):
            e = (
                "Supply curve aggregation found no non-excluded SC points. "
                "Please check your exclusions or subset SC GID selection."
            )
            logger.error(e)
            raise EmptySupplyCurvePointError(e)

        summary = self._summary_to_df(summary)

        return summary

    def run(
        self,
        out_fpath,
        gen_fpath=None,
        args=None,
        max_workers=None,
        sites_per_worker=100,
    ):
        """Run a supply curve aggregation.

        Parameters
        ----------
        gen_fpath : str, optional
            Filepath to HDF5 file with ``reV`` generation output
            results. If ``None``, a simple aggregation without any
            generation, resource, or cost data is performed.

            .. Note:: If executing ``reV`` from the command line, this
              input can be set to ``"PIPELINE"`` to parse the output
              from one of these preceding pipeline steps:
              ``multi-year``, ``collect``, or ``econ``. However, note
              that duplicate executions of any of these commands within
              the pipeline may invalidate this parsing, meaning the
              `gen_fpath` input will have to be specified manually.

            By default, ``None``.
        args : tuple | list, optional
            List of columns to include in summary output table. ``None``
            defaults to all available args defined in the
            :class:`~reV.supply_curve.sc_aggregation.SupplyCurveAggregation`
            documentation. By default, ``None``.
        max_workers : int, optional
            Number of cores to run summary on. ``None`` is all available
            CPUs. By default, ``None``.
        sites_per_worker : int, optional
            Number of sc_points to summarize on each worker.
            By default, ``100``.

        Returns
        -------
        str
            Path to output CSV file containing supply curve aggregation.
        """

        if gen_fpath is None:
            out = Aggregation.run(
                self._excl_fpath,
                self._res_fpath,
                self._tm_dset,
                excl_dict=self._excl_dict,
                resolution=self._resolution,
                excl_area=self._excl_area,
                area_filter_kernel=self._area_filter_kernel,
                min_area=self._min_area,
                pre_extract_inclusions=self._pre_extract_inclusions,
                max_workers=max_workers,
                sites_per_worker=sites_per_worker,
            )
            summary = out["meta"]
        else:
            summary = self.summarize(
                gen_fpath=gen_fpath,
                args=args,
                max_workers=max_workers,
                sites_per_worker=sites_per_worker,
            )

        out_fpath = _format_sc_agg_out_fpath(out_fpath)
        summary.to_csv(out_fpath)

        return out_fpath


def _format_sc_agg_out_fpath(out_fpath):
    """Add CSV file ending and replace underscore, if necessary."""
    if not out_fpath.endswith(".csv"):
        out_fpath = "{}.csv".format(out_fpath)

    project_dir, out_fn = os.path.split(out_fpath)
    out_fn = out_fn.replace(
        "supply_curve_aggregation", "supply-curve-aggregation"
    )
    return os.path.join(project_dir, out_fn)


def _warn_about_large_datasets(gen, dset):
    """Warn user about multi-dimensional datasets in passthrough datasets"""
    dset_shape = gen.shapes.get(dset, (1,))
    if len(dset_shape) > 1:
        msg = (
            "Generation dataset {!r} is not 1-dimensional (shape: {})."
            "You may run into memory errors during aggregation - use "
            "rep-profiles for aggregating higher-order datasets instead!"
            .format(dset, dset_shape)
        )
        logger.warning(msg)
        warn(msg, UserWarning)
